#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

import torch
from classy_vision.losses import LabelSmoothingCrossEntropyLoss, build_loss


class TestLabelSmoothingCrossEntropyLoss(unittest.TestCase):
    def test_build_label_smoothing_cross_entropy(self):
        config = {
            "name": "label_smoothing_cross_entropy",
            "ignore_index": -1,
            "smoothing_param": 0.1,
        }
        crit = build_loss(config)
        self.assertTrue(isinstance(crit, LabelSmoothingCrossEntropyLoss))
        self.assertEqual(crit._ignore_index, -1)

    def test_smoothing_one_hot_targets(self):
        config = {
            "name": "label_smoothing_cross_entropy",
            "ignore_index": -1,
            "smoothing_param": 0.1,
        }
        crit = build_loss(config)
        targets = torch.tensor([[0, 0, 0, 0, 1]])
        self.assertTrue(isinstance(crit, LabelSmoothingCrossEntropyLoss))
        valid_targets = crit.compute_valid_targets(targets, 5)
        self.assertTrue(
            torch.allclose(valid_targets, torch.tensor([[0.0, 0.0, 0.0, 0.0, 1.0]]))
        )
        smoothed_targets = crit.smooth_targets(valid_targets, 5)
        self.assertTrue(
            torch.allclose(
                smoothed_targets,
                torch.tensor([[0.2 / 11, 0.2 / 11, 0.2 / 11, 0.2 / 11, 10.2 / 11]]),
            )
        )

    def test_smoothing_ignore_index_one_hot_targets(self):
        config = {
            "name": "label_smoothing_cross_entropy",
            "ignore_index": -1,
            "smoothing_param": 0.5,
        }
        crit = build_loss(config)
        targets = torch.tensor([[-1, 0, 0, 0, 1]])
        self.assertTrue(isinstance(crit, LabelSmoothingCrossEntropyLoss))
        valid_targets = crit.compute_valid_targets(targets, 5)
        self.assertTrue(
            torch.allclose(valid_targets, torch.tensor([[0.0, 0.0, 0.0, 0.0, 1.0]]))
        )
        smoothed_targets = crit.smooth_targets(valid_targets, 5)
        self.assertTrue(
            torch.allclose(
                smoothed_targets,
                torch.tensor([[1 / 15, 1 / 15, 1 / 15, 1 / 15, 11 / 15]]),
            )
        )

    def test_smoothing_multilabel_one_hot_targets(self):
        config = {
            "name": "label_smoothing_cross_entropy",
            "ignore_index": -1,
            "smoothing_param": 0.5,
        }
        crit = build_loss(config)
        targets = torch.tensor([[1, 0, 0, 0, 1]])
        self.assertTrue(isinstance(crit, LabelSmoothingCrossEntropyLoss))
        valid_targets = crit.compute_valid_targets(targets, 5)
        self.assertTrue(
            torch.allclose(valid_targets, torch.tensor([[1.0, 0.0, 0.0, 0.0, 1.0]]))
        )

        smoothed_targets = crit.smooth_targets(valid_targets, 5)
        self.assertTrue(
            torch.allclose(
                smoothed_targets,
                torch.tensor([[6 / 15, 1 / 15, 1 / 15, 1 / 15, 6 / 15]]),
            )
        )

    def test_smoothing_all_ones_one_hot_targets(self):
        config = {
            "name": "label_smoothing_cross_entropy",
            "ignore_index": -1,
            "smoothing_param": 0.1,
        }
        crit = build_loss(config)
        targets = torch.tensor([[1, 1, 1, 1]])
        self.assertTrue(isinstance(crit, LabelSmoothingCrossEntropyLoss))
        valid_targets = crit.compute_valid_targets(targets, 4)
        self.assertTrue(
            torch.allclose(valid_targets, torch.tensor([[1.0, 1.0, 1.0, 1.0]]))
        )

        smoothed_targets = crit.smooth_targets(valid_targets, 4)
        self.assertTrue(
            torch.allclose(smoothed_targets, torch.tensor([[0.25, 0.25, 0.25, 0.25]]))
        )

    def test_smoothing_mixed_one_hot_targets(self):
        config = {
            "name": "label_smoothing_cross_entropy",
            "ignore_index": -1,
            "smoothing_param": 0.5,
        }
        crit = build_loss(config)
        targets = torch.tensor([[1, 1, 1, 1, 1], [1, 0, 0, 0, 1]])
        self.assertTrue(isinstance(crit, LabelSmoothingCrossEntropyLoss))
        valid_targets = crit.compute_valid_targets(targets, 5)
        self.assertTrue(
            torch.allclose(
                valid_targets,
                torch.tensor([[1.0, 1.0, 1.0, 1.0, 1.0], [1.0, 0.0, 0.0, 0.0, 1.0]]),
            )
        )
        smoothed_targets = crit.smooth_targets(valid_targets, 5)
        self.assertTrue(
            torch.allclose(
                smoothed_targets,
                torch.tensor(
                    [
                        [0.2, 0.2, 0.2, 0.2, 0.2],
                        [6 / 15, 1 / 15, 1 / 15, 1 / 15, 6 / 15],
                    ]
                ),
            )
        )

    def test_smoothing_class_targets(self):
        config = {
            "name": "label_smoothing_cross_entropy",
            "ignore_index": -1,
            "smoothing_param": 0.5,
        }
        crit = build_loss(config)
        targets = torch.tensor([4, -1])
        self.assertTrue(isinstance(crit, LabelSmoothingCrossEntropyLoss))
        valid_targets = crit.compute_valid_targets(targets, 5)
        self.assertTrue(
            torch.allclose(
                valid_targets,
                torch.tensor([[0.0, 0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 0.0, 0.0, 0.0]]),
            )
        )
        smoothed_targets = crit.smooth_targets(valid_targets, 5)
        self.assertTrue(
            torch.allclose(
                smoothed_targets,
                torch.tensor(
                    [
                        [1 / 15, 1 / 15, 1 / 15, 1 / 15, 11 / 15],
                        [0.2, 0.2, 0.2, 0.2, 0.2],
                    ]
                ),
            )
        )

    def test_unnormalized_label_smoothing_cross_entropy(self):
        config = {
            "name": "label_smoothing_cross_entropy",
            "ignore_index": -1,
            "smoothing_param": 0.5,
        }
        crit = LabelSmoothingCrossEntropyLoss.from_config(config)
        outputs = torch.tensor([[0.0, 7.0, 0.0, 0.0, 2.0]])
        targets = torch.tensor([[0, 0, 0, 0, 1]])
        self.assertAlmostEqual(crit(outputs, targets).item(), 5.07609558)

    def test_ignore_index_label_smoothing_cross_entropy(self):
        config = {
            "name": "label_smoothing_cross_entropy",
            "ignore_index": -1,
            "smoothing_param": 0.2,
        }
        crit = LabelSmoothingCrossEntropyLoss.from_config(config)
        outputs = torch.tensor([[0.0, 7.0]])
        targets = torch.tensor([[-1]])
        self.assertAlmostEqual(crit(outputs, targets).item(), 3.50090909)

    def test_class_integer_label_smoothing_cross_entropy(self):
        config = {
            "name": "label_smoothing_cross_entropy",
            "ignore_index": -1,
            "smoothing_param": 0.2,
        }
        crit = LabelSmoothingCrossEntropyLoss.from_config(config)
        outputs = torch.tensor([[1.0, 2.0], [0.0, 2.0]])
        targets = torch.tensor([[0], [1]])
        self.assertAlmostEqual(crit(outputs, targets).item(), 0.76176142)

    def test_deep_copy(self):
        config = {
            "name": "label_smoothing_cross_entropy",
            "ignore_index": -1,
            "smoothing_param": 0.5,
        }
        crit = build_loss(config)
        self.assertTrue(isinstance(crit, LabelSmoothingCrossEntropyLoss))
        outputs = torch.tensor([[0.0, 7.0, 0.0, 0.0, 2.0]])
        targets = torch.tensor([[0, 0, 0, 0, 1]])
        crit(outputs, targets)

        crit2 = copy.deepcopy(crit)
        self.assertAlmostEqual(crit2(outputs, targets).item(), 5.07609558)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch.optim
from classy_vision.generic.util import is_pos_float
from classy_vision.optim.param_scheduler import (
    ClassyParamScheduler,
    build_param_scheduler,
)

from . import ClassyOptimizer, register_optimizer


@register_optimizer("sgd")
class SGD(ClassyOptimizer):
    def __init__(
        self,
        lr_scheduler: ClassyParamScheduler,
        momentum: float = 0,
        weight_decay: float = 0,
        nesterov=False,
    ):
        super().__init__(lr_scheduler=lr_scheduler)

        self.momentum = momentum
        self.weight_decay = weight_decay
        self.nesterov = nesterov

    def init_pytorch_optimizer(self, model):
        super().init_pytorch_optimizer(model)
        self.optimizer = torch.optim.SGD(
            self.param_groups_override,
            lr=self.lr,
            nesterov=self.nesterov,
            momentum=self.momentum,
            weight_decay=self.weight_decay,
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "SGD":
        """Instantiates a SGD from a configuration.

        Args:
            config: A configuration for a SGD.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A SGD instance.
        """
        # Default params
        config["nesterov"] = config.get("nesterov", False)

        assert (
            "lr" in config
        ), "Config must contain a learning rate 'lr' section for SGD optimizer"
        assert (
            "momentum" in config
            and config["momentum"] >= 0.0
            and config["momentum"] < 1.0
            and type(config["momentum"]) == float
        ), "Config must contain a 'momentum' in [0, 1) for SGD optimizer"
        assert "nesterov" in config and isinstance(
            config["nesterov"], bool
        ), "Config must contain a boolean 'nesterov' param for SGD optimizer"
        assert "weight_decay" in config and is_pos_float(
            config["weight_decay"]
        ), "Config must contain a positive 'weight_decay' for SGD optimizer"

        lr_config = config["lr"]
        if not isinstance(lr_config, dict):
            lr_config = {"name": "constant", "value": lr_config}

        lr_config["num_epochs"] = config["num_epochs"]
        lr_scheduler = build_param_scheduler(lr_config)

        return cls(
            lr_scheduler=lr_scheduler,
            momentum=config["momentum"],
            weight_decay=config["weight_decay"],
            nesterov=config["nesterov"],
        )

    @property
    def parameters(self):
        return {
            "lr": self.lr,
            "momentum": self.momentum,
            "weight_decay": self.weight_decay,
            "nesterov": self.nesterov,
        }



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from functools import wraps

import torch


class Arguments(object):
    """Object that looks like input arguments. Used to spoof argparse namespace."""

    def __init__(self, **args):
        self.args = args
        self.__dict__.update(args)

    def __iter__(self):
        return iter(self.args)

    def __eq__(self, other):
        if isinstance(other, Arguments):
            return self.args == other.args
        else:
            return NotImplemented

    def _asdict(self):
        return vars(self)


def skip_if_no_gpu(func):
    """Decorator that can be used to skip GPU tests on non-GPU machines."""
    func.skip_if_no_gpu = True

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not torch.cuda.is_available():
            return
        if torch.cuda.device_count() <= 0:
            return

        return func(*args, **kwargs)

    return wrapper


def repeat_test(original_function=None, *, num_times=3):
    """Decorator that can be used to repeat test multiple times."""

    def repeat_test_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for _ in range(num_times):
                func(*args, **kwargs)

        return wrapper

    # this handles default arguments to decorator:
    if original_function:
        return repeat_test_decorator(original_function)
    return repeat_test_decorator


def make_torch_deterministic(seed=0):
    """Makes Torch code run deterministically."""
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    os.environ["MKL_NUM_THREADS"] = "1"
    os.environ["OMP_NUM_THREADS"] = "1"


def compare_batches(test_fixture, batch1, batch2):
    """Compare two batches. Does not do recursive comparison"""
    test_fixture.assertEqual(type(batch1), type(batch2))
    if isinstance(batch1, (tuple, list)):
        test_fixture.assertEqual(len(batch1), len(batch2))
        for n in range(len(batch1)):
            value1 = batch1[n]
            value2 = batch2[n]
            test_fixture.assertEqual(type(value1), type(value2))
            if torch.is_tensor(value1):
                test_fixture.assertTrue(torch.allclose(value1, value2))
            else:
                test_fixture.assertEqual(value1, value2)

    elif isinstance(batch1, dict):
        test_fixture.assertEqual(batch1.keys(), batch2.keys())
        for key, value1 in batch1.items():
            value2 = batch2[key]
            test_fixture.assertEqual(type(value1), type(value2))
            if torch.is_tensor(value1):
                test_fixture.assertTrue(torch.allclose(value1, value2))
            else:
                test_fixture.assertEqual(value1, value2)


def compare_datasets(test_fixture, dataset1, dataset2):
    test_fixture.assertEqual(len(dataset1), len(dataset2))
    for idx in range(len(dataset1)):
        compare_batches(test_fixture, dataset1[idx], dataset2[idx])


def compare_batchlist_and_dataset_with_skips(
    test_fixture, batch_list, dataset, skip_indices=None
):
    """
    Compares a list of batches and the dataset.  If some samples were
    skipped in the iterator (i.e. if we simulated an error on that
    sample), that should be indicated in the skip_indices list
    """
    if skip_indices is None:
        skip_indices = []
    if isinstance(skip_indices, int):
        skip_indices = [skip_indices]

    skips = 0
    for idx, batch in enumerate(batch_list):
        while (idx + skips) in skip_indices:
            skips += 1
        dataset_batch = dataset[idx + skips]
        compare_batches(test_fixture, batch, dataset_batch)


class MockErrorDataset:
    """
    Dataset used for testing. Wraps a real dataset with a
    batchsize_per_replica, but allows us to delete samples on return
    to simulate errors (similar to what happens with Everstore)
    """

    def __init__(self, dataset):
        self.rebatch_map = {}
        self.dataset = dataset
        self.batchsize_per_replica = dataset.batchsize_per_replica

    def __getitem__(self, idx):
        batch = self.dataset[idx]
        # If rebatch map contains index, resize the batch
        if idx in self.rebatch_map:
            num_samples = self.rebatch_map[idx]
            if num_samples < batch["input"].size()[0]:
                batch["input"] = batch["input"][:num_samples]
                batch["target"] = batch["target"][:num_samples]

        return batch

    def __len__(self):
        return len(self.dataset)


def recursive_unpack(batch):
    """
    Takes a batch of samples, e.g.

      batch = {'input': tensor([256, 3, 224, 224]), 'target': tensor([256])}

    and unpacks them into a list of single samples, e.g.

      [{'input': tensor([1, 3, 224, 224]), 'target': tensor([1])} ... ]
    """
    new_list = []
    if isinstance(batch, dict):
        unpacked_dict = {}
        batchsize_per_replica = -1
        for key, val in batch.items():
            unpacked_dict[key] = recursive_unpack(val)
            batchsize_per_replica = (
                len(unpacked_dict[key])
                if not torch.is_tensor(unpacked_dict[key])
                else 1
            )

        for idx in range(batchsize_per_replica):
            sample = {}
            for key, val in unpacked_dict.items():
                sample[key] = val[idx]

            new_list.append(sample)
        return new_list

    elif isinstance(batch, (list, tuple)):
        unpacked_list = []
        if isinstance(batch, tuple):
            batch = list(batch)

        for val in batch:
            unpacked_list.append(recursive_unpack(val))
            batchsize_per_replica = (
                len(unpacked_list[0]) if not torch.is_tensor(unpacked_list[0]) else 1
            )

        for idx in range(batchsize_per_replica):
            sample = []
            for val in unpacked_list:
                sample.append(val[idx])

            if isinstance(batch, tuple):
                sample = tuple(sample)
            new_list.append(sample)
        return new_list

    elif torch.is_tensor(batch):
        for i in range(batch.size()[0]):
            new_list.append(batch[i])
        return new_list

    raise TypeError("Unexpected type %s passed to unpack" % type(batch))


def compare_model_state(test_fixture, state, state2, check_heads=True):
    for k in state["model"]["trunk"].keys():
        if not torch.allclose(state["model"]["trunk"][k], state2["model"]["trunk"][k]):
            print(k, state["model"]["trunk"][k], state2["model"]["trunk"][k])
        test_fixture.assertTrue(
            torch.allclose(state["model"]["trunk"][k], state2["model"]["trunk"][k])
        )
    if check_heads:
        for block, head_states in state["model"]["heads"].items():
            for head_id, states in head_states.items():
                for k in states.keys():
                    test_fixture.assertTrue(
                        torch.allclose(
                            state["model"]["heads"][block][head_id][k],
                            state2["model"]["heads"][block][head_id][k],
                        )
                    )


def compare_samples(test_fixture, sample1, sample2):
    test_fixture.assertEqual(sample1.keys(), sample2.keys())
    test_fixture.assertTrue(torch.is_tensor(sample1["input"]))
    test_fixture.assertTrue(torch.is_tensor(sample2["input"]))
    test_fixture.assertTrue(torch.is_tensor(sample1["target"]))
    test_fixture.assertTrue(torch.is_tensor(sample2["target"]))

    test_fixture.assertTrue(torch.allclose(sample1["input"], sample2["input"]))
    test_fixture.assertTrue(torch.allclose(sample1["target"], sample2["target"]))


def compare_states(test_fixture, state_1, state_2, check_heads=True):
    """
    Tests the classy state dicts for equality, but skips the member objects
    which implement their own {get, set}_classy_state functions.
    """
    # check base_model
    compare_model_state(
        test_fixture, state_1["base_model"], state_2["base_model"], check_heads
    )
    # check losses
    test_fixture.assertEqual(len(state_1["losses"]), len(state_2["losses"]))
    for loss_1, loss_2 in zip(state_1["losses"], state_2["losses"]):
        test_fixture.assertAlmostEqual(loss_1, loss_2)

    for key in ["base_model", "meters", "optimizer", "losses"]:
        # we trust that these have been tested using their unit tests or
        # by the code above
        test_fixture.assertIn(key, state_1)
        test_fixture.assertIn(key, state_2)
        del state_1[key]
        del state_2[key]
    test_fixture.assertDictEqual(state_1, state_2)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

from classy_vision.dataset.core.random_image_datasets import (
    RandomImageBinaryClassDataset,
)
from classy_vision.dataset.transforms.util import build_field_transform_default_imagenet


class LightingTransformTest(unittest.TestCase):
    def get_test_image_dataset(self):
        return RandomImageBinaryClassDataset(
            crop_size=224, class_ratio=0.5, num_samples=100, seed=0
        )

    def test_lighting_transform_no_errors(self):
        """
        Tests that the lighting transform runs without any errors.
        """
        dataset = self.get_test_image_dataset()

        config = [{"name": "ToTensor"}, {"name": "lighting"}]
        transform = build_field_transform_default_imagenet(config)
        sample = dataset[0]
        try:
            # test that lighting has been registered and runs without errors
            transform(sample)
        except Exception:
            self.fail("LightingTransform raised an exception")
        return



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from test.generic.meter_test_utils import ClassificationMeterTest

import torch
from classy_vision import meters
from classy_vision.meters import RecallAtKMeter


class TestRecallAtKMeter(ClassificationMeterTest):
    def test_recall_meter_registry(self):
        meter = meters.build_meter({"name": "recall_at_k", "topk": [1, 3]})
        self.assertTrue(isinstance(meter, RecallAtKMeter))

    def test_single_meter_update_and_reset(self):
        """
        This test verifies that the meter works as expected on a single
        update + reset + same single update.
        """
        meter = RecallAtKMeter(topk=[1, 2])

        # Batchsize = 3, num classes = 3, score is probability of class
        model_output = torch.tensor(
            [
                [0.2, 0.4, 0.4],  # top-1: 1/2, top-2: 1/2
                [0.2, 0.65, 0.15],  # top-1: 1, top-2: 1/0
                [0.33, 0.33, 0.34],  # top-1: 2, top-2: 2/0?1
            ]
        )

        # One-hot encoding, 1 = positive for class
        # sample-1: 1, sample-2: 0, sample-3: 0,1,2
        target = torch.tensor([[0, 1, 0], [1, 0, 0], [1, 1, 1]])

        # Note for ties, we select randomly, so we should not use ambiguous ties
        expected_value = {"top_1": 2 / 5.0, "top_2": 4 / 5.0}

        self.meter_update_and_reset_test(meter, model_output, target, expected_value)

    def test_double_meter_update_and_reset(self):
        meter = RecallAtKMeter(topk=[1, 2])

        # Batchsize = 3, num classes = 3, score is probability of class
        model_outputs = [
            torch.tensor([[0.3, 0.4, 0.3], [0.2, 0.65, 0.15], [0.33, 0.33, 0.34]]),
            torch.tensor([[0.05, 0.4, 0.05], [0.15, 0.65, 0.2], [0.4, 0.2, 0.4]]),
        ]

        # One-hot encoding, 1 = positive for class
        # batch-1: sample-1: 1, sample-2: 0, sample-3: 0,1,2
        # batch-2: sample-1: 1, sample-2: 1, sample-3: 1
        targets = [
            torch.tensor([[0, 1, 0], [1, 0, 0], [1, 1, 1]]),
            torch.tensor([[0, 1, 0], [0, 1, 0], [0, 1, 0]]),
        ]

        # First batch has top-1 recall of 2/5.0, top-2 recall of 4/5.0
        # Second batch has top-1 recall of 2/3.0, top-2 recall of 2/3.0
        expected_value = {"top_1": 4 / 8.0, "top_2": 6 / 8.0}

        self.meter_update_and_reset_test(meter, model_outputs, targets, expected_value)

    def test_meter_invalid_model_output(self):
        meter = RecallAtKMeter(topk=[1, 2])
        # This model output has 3 dimensions instead of expected 2
        model_output = torch.tensor(
            [[[0.33, 0.33, 0.34], [1, 2, 3]], [[-1, -3, -4], [-10, -90, -100]]]
        )
        target = torch.tensor([[0, 1, 0], [1, 0, 0], [1, 1, 1]])

        self.meter_invalid_meter_input_test(meter, model_output, target)

    def test_meter_invalid_target(self):
        meter = RecallAtKMeter(topk=[1, 2])
        model_output = torch.tensor(
            [
                [0.2, 0.4, 0.4],  # top-1: 1/2, top-2: 1/2
                [0.2, 0.65, 0.15],  # top-1: 1, top-2: 1/0
                [0.33, 0.33, 0.34],  # top-1: 2, top-2: 2/0/1
            ]
        )
        # Target shape does not match model shape
        target = torch.tensor([0, 1, 2])

        self.meter_invalid_meter_input_test(meter, model_output, target)

    def test_meter_invalid_topk(self):
        meter = RecallAtKMeter(topk=[1, 5])
        model_output = torch.tensor(
            [
                [0.2, 0.4, 0.4],  # top-1: 1/2, top-2: 1/2
                [0.2, 0.65, 0.15],  # top-1: 1, top-2: 1/0
                [0.33, 0.33, 0.34],  # top-1: 2, top-2: 2/0/1
            ]
        )
        target = torch.tensor([[0, 1, 0], [1, 0, 0], [1, 1, 1]])

        self.meter_invalid_meter_input_test(meter, model_output, target)

    def test_meter_get_set_classy_state_test(self):
        # In this test we update meter0 with model_output0 & target0
        # and we update meter1 with model_output1 & target1 then
        # transfer the state from meter1 to meter0 and validate they
        # give same expected value.
        #
        # Expected value is the expected value of meter1 For this test
        # to work, top-1 / top-2 values of meter0 / meter1 should be
        # different
        meters = [RecallAtKMeter(topk=[1, 2]), RecallAtKMeter(topk=[1, 2])]
        model_outputs = [
            torch.tensor([[0.05, 0.4, 0.05], [0.2, 0.65, 0.15], [0.33, 0.33, 0.34]]),
            torch.tensor([[0.05, 0.4, 0.05], [0.15, 0.65, 0.2], [0.4, 0.2, 0.4]]),
        ]
        targets = [
            torch.tensor([[0, 1, 0], [1, 0, 0], [1, 1, 0]]),
            torch.tensor([[0, 1, 0], [0, 1, 0], [0, 1, 0]]),
        ]

        # Second update's expected value
        expected_value = {"top_1": 2 / 3.0, "top_2": 2 / 3.0}

        self.meter_get_set_classy_state_test(
            meters, model_outputs, targets, expected_value
        )

    def test_meter_distributed(self):
        # Meter0 will execute on one process, Meter1 on the other
        meters = [RecallAtKMeter(topk=[1, 2]), RecallAtKMeter(topk=[1, 2])]

        # Batchsize = 3, num classes = 3, score is probability of class
        model_outputs = [
            torch.tensor(
                [[0.3, 0.4, 0.3], [0.2, 0.65, 0.15], [0.33, 0.33, 0.34]]
            ),  # Meter 0
            torch.tensor(
                [[0.05, 0.4, 0.05], [0.15, 0.65, 0.2], [0.4, 0.2, 0.4]]
            ),  # Meter 1
            torch.tensor(
                [[0.3, 0.4, 0.3], [0.2, 0.65, 0.15], [0.33, 0.33, 0.34]]
            ),  # Meter 0
            torch.tensor(
                [[0.05, 0.4, 0.05], [0.15, 0.65, 0.2], [0.4, 0.2, 0.4]]
            ),  # Meter 1
        ]

        # Class 0 is the correct class for sample 1, class 2 for sample 2, etc
        targets = [
            torch.tensor([[0, 1, 0], [1, 0, 0], [1, 1, 1]]),  # Meter 0
            torch.tensor([[0, 1, 0], [0, 1, 0], [0, 1, 0]]),  # Meter 1
            torch.tensor([[0, 1, 0], [1, 0, 0], [1, 1, 1]]),  # Meter 0
            torch.tensor([[0, 1, 0], [0, 1, 0], [0, 1, 0]]),  # Meter 1
        ]

        # In first two updates there are 4 correct top-1 out of 8
        # total, 6 correct in top 2 out of 8.  The same occurs in the
        # second two updates and is added to first
        expected_values = [
            {"top_1": 4 / 8.0, "top_2": 6 / 8.0},  # After one update to each meter
            {"top_1": 8 / 16.0, "top_2": 12 / 16.0},  # After two updates to each meter
        ]

        self.meter_distributed_test(meters, model_outputs, targets, expected_values)

    def test_non_onehot_target(self):
        """
        This test verifies that the meter works as expected on a single
        update + reset + same single update.
        """
        meter = RecallAtKMeter(topk=[1, 2], target_is_one_hot=False, num_classes=3)

        # Batchsize = 2, num classes = 3, score is probability of class
        model_outputs = [
            torch.tensor([[0.05, 0.4, 0.05], [0.15, 0.65, 0.2], [0.4, 0.2, 0.4]]),
            torch.tensor([[0.2, 0.4, 0.4], [0.2, 0.65, 0.15], [0.1, 0.8, 0.1]]),
        ]

        # One-hot encoding, 1 = positive for class
        targets = [
            torch.tensor([[1], [1], [1]]),  # [[0, 1, 0], [0, 1, 0], [0, 1, 0]]
            torch.tensor([[0], [1], [2]]),  # [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        ]

        # Note for ties, we select randomly, so we should not use ambiguous ties
        # First batch has top-1 recall of 2/3.0, top-2 recall of 2/6.0
        # Second batch has top-1 recall of 1/3.0, top-2 recall of 4/6.0
        expected_value = {"top_1": 3 / 6.0, "top_2": 6 / 12.0}

        self.meter_update_and_reset_test(meter, model_outputs, targets, expected_value)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Callable, Dict, Optional, Sequence, Union

from classy_vision.dataset.transforms import ClassyTransform
from classy_vision.generic.distributed_util import get_rank, get_world_size
from classy_vision.generic.util import is_pos_int
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler


def _return_true(_sample):
    return True


class ClassyDataset:
    """
    Class representing a dataset abstraction.

    This class wraps a :class:`torch.utils.data.Dataset` via the `dataset` attribute
    and configures the dataloaders needed to access the datasets.
    Transforms which need to be applied to the data should be specified in this class.
    ClassyDataset can be used to instantiate datasets from a configuration file as well.
    """

    def __init__(
        self,
        dataset: Sequence,
        split: Optional[str],
        batchsize_per_replica: int,
        shuffle: bool,
        transform: Optional[Union[ClassyTransform, Callable]],
        num_samples: Optional[int],
    ) -> None:
        """
        Constructor for a ClassyDataset.

        Args:
            split: Split of dataset to use ("train", "test")
            batchsize_per_replica: Positive integer indicating batch size for each
                replica
            shuffle: Whether we should shuffle between epochs
            transform: Transform to be applied to each sample
            num_samples: When set, this restricts the number of samples provided by
                the dataset
        """
        # Asserts:
        assert is_pos_int(
            batchsize_per_replica
        ), "batchsize_per_replica must be a positive int"
        assert isinstance(shuffle, bool), "shuffle must be a boolean"
        assert num_samples is None or is_pos_int(
            num_samples
        ), "num_samples must be a positive int or None"

        # Assignments:
        self.split = split
        self.batchsize_per_replica = batchsize_per_replica
        self.shuffle = shuffle
        self.transform = transform
        self.num_samples = num_samples
        self.dataset = dataset

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ClassyDataset":
        """Instantiates a ClassyDataset from a configuration.

        Args:
            config: A configuration for the ClassyDataset.

        Returns:
            A ClassyDataset instance.
        """
        raise NotImplementedError

    @classmethod
    def parse_config(cls, config: Dict[str, Any]):
        """
        This function parses out common config options.

        Args:
            config: A dict with the following string keys -

                | batchsize_per_replica (int): Must be a positive int, batch size
                |    for each replica
                | use_shuffle (bool): Whether to enable shuffling for the dataset
                | num_samples (int, optional): When set, restricts the number of samples
                |    in a dataset
                | transforms: list of tranform configurations to be applied in order

        Returns:
            A tuple containing the following variables -
                | transform_config: Config for the dataset transform. Can be passed to
                |    :func:`build_transform`
                | batchsize_per_replica: Batch size per replica
                | shuffle: Whether we should shuffle between epochs
                | num_samples: When set, restricts the number of samples in a dataset
        """
        batchsize_per_replica = config.get("batchsize_per_replica")
        shuffle = config.get("use_shuffle")
        num_samples = config.get("num_samples")
        transform_config = config.get("transforms")
        return transform_config, batchsize_per_replica, shuffle, num_samples

    def __getitem__(self, idx: int):
        assert idx >= 0 and idx < len(
            self.dataset
        ), "Provided idx is outside of dataset range"
        sample = self.dataset[idx]
        if self.transform is None:
            return sample
        return self.transform(sample)

    def __len__(self):
        assert self.num_samples is None or self.num_samples <= len(
            self.dataset
        ), "Num samples mus be less than length of base dataset"
        return len(self.dataset) if self.num_samples is None else self.num_samples

    def _get_sampler(self, epoch: int):
        """
        Return a :class:`torch.utils.data.sampler.Sampler` to sample the data.

        This is used to distribute the data across the replicas. If shuffling
        is enabled, every epoch will have a different shuffle.

        Args:
            epoch: The epoch being fetched.

        Returns:
            A sampler which tells the data loader which sample to load next.
        """
        world_size = get_world_size()
        rank = get_rank()
        sampler = DistributedSampler(
            self, num_replicas=world_size, rank=rank, shuffle=self.shuffle
        )
        sampler.set_epoch(epoch)
        return sampler

    def iterator(self, *args, **kwargs):
        """
        Returns an iterable which can be used to iterate over the data.

        Args:
            shuffle_seed (int, optional): Seed for the shuffle
            current_phase_id (int, optional): The epoch being fetched. Needed so that
                each epoch has a different shuffle order
        Returns:
            An iterable over the data
        """
        # TODO: Fix naming to be consistent (i.e. everyone uses epoch)
        shuffle_seed = kwargs.get("shuffle_seed", 0)
        assert isinstance(shuffle_seed, int), "Shuffle seed must be an int"
        epoch = kwargs.get("current_phase_id", 0)
        assert isinstance(epoch, int), "Epoch must be an int"

        offset_epoch = shuffle_seed + epoch

        return DataLoader(
            self,
            batch_size=self.batchsize_per_replica,
            num_workers=kwargs.get("num_workers", 0),
            pin_memory=kwargs.get("pin_memory", False),
            multiprocessing_context=kwargs.get("multiprocessing_context", None),
            sampler=self._get_sampler(epoch=offset_epoch),
        )

    def get_batchsize_per_replica(self):
        """
        Get the batch size per replica.

        Returns:
            The batch size for each replica.
        """
        return self.batchsize_per_replica

    def get_global_batchsize(self):
        """
        Get the global batch size, combined over all the replicas.

        Returns:
            The overall batch size of the dataset.
        """
        return self.get_batchsize_per_replica() * get_world_size()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch.nn as nn
from classy_vision.generic.util import is_pos_int
from classy_vision.heads import ClassyHead, register_head


@register_head("fully_connected")
class FullyConnectedHead(ClassyHead):
    """This head defines a 2d average pooling layer
    (:class:`torch.nn.AdaptiveAvgPool2d`) followed by a fully connected
    layer (:class:`torch.nn.Linear`).
    """

    def __init__(self, unique_id: str, num_classes: int, in_plane: int):
        """Constructor for FullyConnectedHead

        Args:
            unique_id: A unique identifier for the head. Multiple instances of
                the same head might be attached to a model, and unique_id is used
                to refer to them.

            num_classes: Number of classes for the head. If None, then the fully
                connected layer is not applied.

            in_plane: Input size for the fully connected layer.
        """
        super().__init__(unique_id, num_classes)
        assert num_classes is None or is_pos_int(num_classes)
        assert is_pos_int(in_plane)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = None if num_classes is None else nn.Linear(in_plane, num_classes)

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "FullyConnectedHead":
        """Instantiates a FullyConnectedHead from a configuration.

        Args:
            config: A configuration for a FullyConnectedHead.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A FullyConnectedHead instance.
        """
        num_classes = config.get("num_classes", None)
        in_plane = config["in_plane"]
        return cls(config["unique_id"], num_classes, in_plane)

    def forward(self, x):
        # perform average pooling:
        out = self.avgpool(x)

        # final classifier:
        out = out.reshape(out.size(0), -1)
        if self.fc is not None:
            out = self.fc(out)
        return out



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Callable, Dict, Optional

import torch
from classy_vision.models import ClassyModel

from .param_scheduler.classy_vision_param_scheduler import (
    ClassyParamScheduler,
    UpdateInterval,
)


class ClassyOptimizer:
    """
    Base class for classy optimizers.

    This wraps a :class:`torch.optim.Optimizer` instance, handles learning
    rate scheduling by using a :class:`ClassyParamScheduler` and supports specifying
    regularized and unregularized param groups. Specifying unregularized params is
    especially useful to avoid applying weight decay on batch norm. See
    :method:ClassyModel.get_optimizer_params for more information.

    Deriving classes can extend functionality be overriding the appropriate functions.
    """

    def __init__(self, lr_scheduler: ClassyParamScheduler):
        """
        Constructor for ClassyOptimizer.

        Args:
            lr_scheduler: The learning rate scheduler to use.
        """
        self.lr_scheduler = lr_scheduler
        self.lr = self.lr_scheduler(0)
        self.optimizer = None
        self.optimizer_params = None

    def _validate_and_get_optimizer_params(self, model: ClassyModel) -> Dict[str, Any]:
        """
        Validate and return the optimizer params.

        The optimizer params are fetched from
        :method:`ClassyVision.get_optimizer_params`.

        Args:
            model: The model to get the params from.

        Returns:
            A dict containing "regularized_params" and "unregularized_params".
            Weight decay will only be applied to "regularized_params".
        """
        if isinstance(model, torch.nn.parallel.DistributedDataParallel):
            optimizer_params = model.module.get_optimizer_params()
        else:
            optimizer_params = model.get_optimizer_params()

        assert isinstance(optimizer_params, dict) and set(optimizer_params.keys()) == {
            "regularized_params",
            "unregularized_params",
        }, "get_optimizer_params() of {0} should return dict with exact two keys\
            'regularized_params', 'unregularized_params'".format(
            type(model).__name__
        )

        trainable_params = [
            params for params in model.parameters() if params.requires_grad
        ]
        assert len(trainable_params) == len(
            optimizer_params["regularized_params"]
        ) + len(optimizer_params["unregularized_params"]), (
            "get_optimizer_params() of {0} should return params that cover all"
            "trainable params of model".format(type(model).__name__)
        )

        return optimizer_params

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ClassyOptimizer":
        """Instantiates a ClassyOptimizer from a configuration.

        Args:
            config: A configuration for the ClassyOptimizer.

        Returns:
            A ClassyOptimizer instance.
        """
        raise NotImplementedError

    @property
    def parameters(self) -> Dict[str, Any]:
        """
        Get the parameters of the optimizer which need to be overridden. All optimizer
        param groups will use these parameters.

        Returns:
            A kwarg dictionary that will be used to override optimizer args.
        """
        return {"lr": self.lr}

    def init_pytorch_optimizer(self, model: ClassyModel) -> None:
        """
        Initialize the underlying :class:`torch.optim.Optimizer` instance.

        Using the provided model, create param groups for the optimizer with a
        weight decay override for params which should be left unregularized.

        Note:
            Deriving classes should initialize the underlying Pytorch optimizer
            in this call. The simplest way to do this after a call to
            super().init_pytorch_optimizer().

        Warning:
            This should called only after the model has been moved to the correct
            device.
        """
        self.optimizer_params = self._validate_and_get_optimizer_params(model)

        param_groups_override = []
        self.contains_unregularized_params = False
        if len(self.optimizer_params["unregularized_params"]) != 0:
            param_groups_override.append(
                {
                    "params": self.optimizer_params["unregularized_params"],
                    "weight_decay": 0.0,
                }
            )
            self.contains_unregularized_params = True

        if len(self.optimizer_params["regularized_params"]) != 0:
            param_groups_override.append(
                {"params": self.optimizer_params["regularized_params"]}
            )
        self.param_groups_override = param_groups_override

    def get_classy_state(self) -> Dict[str, Any]:
        """Get the state of the ClassyOptimizer.

        The returned state is used for checkpointing.

        Returns:
            A state dictionary containing the state of the optimizer.
        """
        return {"optim": self.optimizer.state_dict(), "parameters": self.parameters}

    def set_classy_state(self, state: Dict[str, Any]) -> None:
        """Set the state of the ClassyOptimizer.

        Args:
            state_dict: The state dictionary. Must be the output of a call to
                :method:`get_classy_state`.

        This is used to load the state of the optimizer from a checkpoint.
        """
        self.optimizer.load_state_dict(state["optim"])
        for param_name, param_value in state["parameters"].items():
            setattr(self, param_name, param_value)

    def backward(self, loss: torch.Tensor) -> None:
        """
        Computer gradients with respect to the loss.

        Calls :method:`zero_grad` and then computes the gradient using
        :func:`torch.Tensor.backward`. See :mod:`torch.autograd` for more information.
        """
        # TODO (aadcock): Add gradient accumulation logic
        self.zero_grad()
        loss.backward()

    def update_schedule_on_epoch(self, where: float) -> None:
        """
        Update the param schedule at the end of an epoch.

        This should be called by the task at the end of every epoch to update the
        schedule of epoch based param schedulers (See :class:`ClassyParamScheduler` for
        more information).

        Args:
            where: where we are in terms of training progress (output of
                :method:`ClassyTask.where`)
        """
        assert self.lr_scheduler.update_interval in [
            UpdateInterval.EPOCH,
            UpdateInterval.STEP,
        ]

        if self.lr_scheduler.update_interval == UpdateInterval.EPOCH:
            self._update_schedule(where)

    def update_schedule_on_step(self, where: float) -> None:
        """
        Update the param schedule at the end of a train step.

        This should be called by the task at the end of every train step (
        :method:`ClassyTask.train_step`) to update the schedule of step based param
        schedulers (See :class:`ClassyParamScheduler` for more information).

        Args:
            where: where we are in terms of training progress (output of
                :method:`ClassyTask.where`)
        """
        assert self.lr_scheduler.update_interval in [
            UpdateInterval.EPOCH,
            UpdateInterval.STEP,
        ]

        if self.lr_scheduler.update_interval == UpdateInterval.STEP:
            self._update_schedule(where)

    def _update_schedule(self, where: float) -> None:
        """
        Args:
            where: where we are in terms of training progress (output of
                :method:`ClassyTask.where`)
        """
        self.lr = self.lr_scheduler(where)
        for group in self.optimizer.param_groups:
            group.update(self.parameters)

        # Here there's an assumption that pytorch optimizer maintain the order of
        # param_groups and batch_norm param_group is 0th param_group as initially
        # set in the __init__ call.
        # It seems like pytorch optim doesn't have way to get params by 'id':
        #   See thread https://github.com/pytorch/pytorch/issues/1489
        if self.contains_unregularized_params:
            self.optimizer.param_groups[0].update(weight_decay=0.0)

    def step(self, closure: Optional[Callable] = None):
        """
        Performs a single optimization step.

        See :method:`torch.optim.Optimizer.step` for more information.

        Args:
            closure: A closure that re-evaluates the model and returns the loss
        """
        self.optimizer.step(closure)

    def zero_grad(self):
        """
        Clears the gradients of all optimized parameters.

        See :method:`torch.optim.Optimizer.zero_grad` for more information.
        """
        self.optimizer.zero_grad()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import typing
import unittest
import unittest.mock as mock
from pathlib import Path
from test.generic.config_utils import get_fast_test_task_config, get_test_model_configs
from test.generic.utils import compare_model_state, compare_states

import classy_vision.generic.util as util
import torch
from classy_vision.generic.util import update_classy_model, update_classy_state
from classy_vision.models import build_model
from classy_vision.tasks import build_task
from classy_vision.trainer import LocalTrainer


ROOT = Path(__file__).parent


@mock.patch("torch.tensor")
def get_mock_tensor(mock_class):
    def get_cuda_tensor():
        t = torch.tensor([1, 2, 3])
        t.is_cuda = True
        return t

    mock_class.return_value.cuda.return_value = get_cuda_tensor()
    mock_class.is_cuda = False
    return torch.tensor([1, 2, 3])


class TestUtilMethods(unittest.TestCase):
    def _get_base_pred_probs(self):
        return torch.tensor(
            [
                [0.92, 0.08],  # 1
                [0.91, 0.09],  # 0
                [0.89, 0.11],  # 0
                [0.79, 0.21],  # 0
                [0.78, 0.22],  # 0
                [0.69, 0.31],  # 1
                [0.68, 0.32],  # 0
                [0.59, 0.41],  # 1
                [0.58, 0.42],  # 0
                [0.49, 0.51],  # 1
                [0.48, 0.52],  # 1
                [0.39, 0.61],  # 0
                [0.38, 0.62],  # 0
                [0.29, 0.71],  # 1
                [0.28, 0.72],  # 1
                [0.19, 0.81],  # 0
                [0.18, 0.82],  # 1
                [0.09, 0.91],  # 1
                [0.08, 0.92],  # 1
                [0.07, 0.93],  # 0
                [0.06, 0.94],  # 1
                [0.03, 0.97],  # 0
            ]
        )

    def _get_base_targets(self):
        return torch.tensor(
            [1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 0]
        ).int()

    def _get_base_class_hist(self):
        return torch.stack(
            [
                torch.Tensor(
                    [1, 1, 0, 1, 0, 0, 0, 2, 0, 0, 0, 1, 0, 1, 0, 2, 0, 1, 1, 0]
                ),
                torch.Tensor(
                    [0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 2, 0, 0, 0, 2, 0, 1, 0, 3, 0]
                ),
            ],
            dim=1,
        ).long()

    def _get_base_total_hist(self):
        return torch.stack(
            [
                torch.Tensor(
                    [1, 4, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 1, 2, 0]
                ),
                torch.Tensor(
                    [0, 2, 1, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 4, 1]
                ),
            ],
            dim=1,
        ).long()

    def test_create_class_histograms_success(self):
        pred_probs = self._get_base_pred_probs()
        targets = self._get_base_targets()

        class_hist, total_hist = util.create_class_histograms(pred_probs, targets, 20)
        torch.testing.assert_allclose(class_hist, self._get_base_class_hist())
        torch.testing.assert_allclose(total_hist, self._get_base_total_hist())

    def test_create_class_histograms_fail(self):
        pred_probs = self._get_base_pred_probs()
        targets = self._get_base_targets()

        # Torch tensors only
        with self.assertRaises(AssertionError):
            class_hist, total_hist = util.create_class_histograms(
                pred_probs.numpy(), targets, 20
            )

        # Torch tensors only
        with self.assertRaises(AssertionError):
            class_hist, total_hist = util.create_class_histograms(
                pred_probs, targets.numpy(), 20
            )

        # Prediction and target are same size
        with self.assertRaises(AssertionError):
            class_hist, total_hist = util.create_class_histograms(
                pred_probs[0:5, :], targets, 20
            )

        # Prediction is between 0 and 1
        with self.assertRaises(AssertionError):
            pred_probs[0, :] = torch.tensor([-0.1, 1.1])
            class_hist, total_hist = util.create_class_histograms(
                pred_probs, targets, 20
            )

    def test_compute_pr_curves(self):
        class_hist = self._get_base_class_hist()
        total_hist = self._get_base_total_hist()

        pr_curves = util.compute_pr_curves(class_hist, total_hist)
        # For curves without duplicates removed / precisions cleaned
        # up, see: P60302268
        exp_pos_prec = torch.tensor(
            [
                3.0 / 5.0,
                4.0 / 7.0,
                6.0 / 9.0,
                6.0 / 11.0,
                8.0 / 13.0,
                9.0 / 15.0,
                10.0 / 17.0,
                10.0 / 19.0,
                10.0 / 20.0,
                11.0 / 22.0,
            ],
            dtype=torch.double,
        )
        exp_pos_recall = torch.tensor(
            [
                3.0 / 11.0,
                4.0 / 11.0,
                6.0 / 11.0,
                6.0 / 11.0,
                8.0 / 11.0,
                9.0 / 11.0,
                10.0 / 11.0,
                10.0 / 11.0,
                10.0 / 11.0,
                11.0 / 11.0,
            ],
            dtype=torch.double,
        )

        exp_neg_prec = torch.tensor(
            [
                1.0 / 2.0,
                2.0 / 3.0,
                4.0 / 5.0,
                5.0 / 7.0,
                6.0 / 9.0,
                6.0 / 11.0,
                8.0 / 13.0,
                8.0 / 15.0,
                9.0 / 17.0,
                10.0 / 21.0,
                11.0 / 22.0,
            ],
            dtype=torch.double,
        )
        exp_neg_recall = torch.tensor(
            [
                1.0 / 11.0,
                2.0 / 11.0,
                4.0 / 11.0,
                5.0 / 11.0,
                6.0 / 11.0,
                6.0 / 11.0,
                8.0 / 11.0,
                8.0 / 11.0,
                9.0 / 11.0,
                10.0 / 11.0,
                11.0 / 11.0,
            ],
            dtype=torch.double,
        )

        torch.testing.assert_allclose(pr_curves["prec"][1], exp_pos_prec)
        torch.testing.assert_allclose(pr_curves["prec"][0], exp_neg_prec)

        torch.testing.assert_allclose(pr_curves["recall"][1], exp_pos_recall)
        torch.testing.assert_allclose(pr_curves["recall"][0], exp_neg_recall)

        torch.testing.assert_allclose(
            pr_curves["ap"][1], torch.tensor(0.589678058127256).double()
        )
        torch.testing.assert_allclose(
            pr_curves["ap"][0], torch.tensor(0.6073388287292031).double()
        )

    def test_compute_pr_curves_fail(self):
        class_hist = self._get_base_class_hist()
        total_hist = self._get_base_total_hist()

        # invalid histograms
        with self.assertRaises(AssertionError):
            class_hist += torch.ones(class_hist.size(), dtype=torch.int64) * 100
            util.compute_pr_curves(class_hist, total_hist)

        # Doesn't accept numpy
        with self.assertRaises(AssertionError):
            util.compute_pr_curves(class_hist.numpy(), total_hist)

        with self.assertRaises(AssertionError):
            util.compute_pr_curves(class_hist, total_hist.numpy())

        # Longs only
        with self.assertRaises(AssertionError):
            util.compute_pr_curves(class_hist.float(), total_hist.float())

        # Bad tensor size
        with self.assertRaises(AssertionError):
            util.compute_pr_curves(class_hist.view(40, 1), total_hist)

        with self.assertRaises(AssertionError):
            util.compute_pr_curves(class_hist, total_hist.view(40, 1))

    def test_recursive_copy_to_gpu(self):
        tensor_a = get_mock_tensor()
        tensor_b = get_mock_tensor()

        valid_gpu_copy_value = tensor_a
        gpu_value = util.recursive_copy_to_gpu(valid_gpu_copy_value)
        self.assertTrue(gpu_value.is_cuda)

        valid_recursive_copy_value = [[tensor_a]]
        gpu_value = util.recursive_copy_to_gpu(valid_recursive_copy_value)
        self.assertTrue(gpu_value[0][0].is_cuda)

        valid_gpu_copy_collections = [
            (tensor_a, tensor_b),
            [tensor_a, tensor_b],
            {"tensor_a": tensor_a, "tensor_b": tensor_b},
        ]
        for value in valid_gpu_copy_collections:
            gpu_value = util.recursive_copy_to_gpu(value)
            if isinstance(value, dict):
                self.assertTrue(gpu_value["tensor_a"].is_cuda)
                self.assertTrue(gpu_value["tensor_b"].is_cuda)
            else:
                self.assertEqual(len(gpu_value), 2)
                self.assertTrue(gpu_value[0].is_cuda)
                self.assertTrue(gpu_value[1].is_cuda)

        invalid_gpu_copy_values = [1234, True, 1.0]
        for value in invalid_gpu_copy_values:
            with self.assertRaises(AttributeError):
                gpu_value = util.recursive_copy_to_gpu(value)

        invalid_gpu_copy_depth = [
            ((((tensor_a, tensor_b), tensor_b), tensor_b), tensor_b),
            {"tensor_map_a": {"tensor_map_b": {"tensor_map_c": {"tensor": tensor_a}}}},
            [[[[tensor_a, tensor_b], tensor_b], tensor_b], tensor_b],
            "abcd",  # Strings are sequences, includeing single char strings
        ]
        for value in invalid_gpu_copy_depth:
            with self.assertRaises(ValueError):
                gpu_value = util.recursive_copy_to_gpu(value, max_depth=3)

    _json_config_file = ROOT / "generic_util_json_blob_test.json"

    def _get_config(self):
        return {
            "name": "test_task",
            "num_epochs": 12,
            "loss": {"name": "test_loss"},
            "dataset": {
                "name": "test_data",
                "batchsize_per_replica": 8,
                "use_pairs": False,
                "num_samples": None,
                "use_shuffle": {"train": True, "test": False},
            },
            "meters": [{"name": "test_meter", "test_param": 0.1}],
            "model": {"name": "test_model", "architecture": [1, 2, 3, 4]},
            "optimizer": {
                "name": "test_optimizer",
                "test_param": {
                    "name": "test_scheduler",
                    "values": [0.1, 0.01, 0.001, 0.0001],
                },
            },
        }

    def test_load_config(self):
        expected_config = self._get_config()
        config = util.load_json(self._json_config_file)

        self.assertEqual(config, expected_config)

    def test_torch_seed(self):
        # test that using util.torch_seed doesn't impact the generation of
        # random numbers outside its context and that random numbers generated
        # within its context are the same as setting a manual seed
        torch.manual_seed(0)
        torch.randn(10)
        random_tensor_1 = torch.randn(10)
        torch.manual_seed(0)
        torch.randn(10)
        with util.torch_seed(1):
            random_tensor_2 = torch.randn(10)
        self.assertTrue(torch.equal(torch.randn(10), random_tensor_1))
        torch.manual_seed(1)
        self.assertTrue(torch.equal(torch.randn(10), random_tensor_2))

    def test_get_model_dummy_input(self):
        for config in get_test_model_configs():
            model = build_model(config)  # pass in a dummy model for the cuda check
            batchsize = 8
            # input_key is list
            input_key = ["audio", "video"]
            input_shape = [[3, 40, 100], [4, 16, 223, 223]]  # dummy input shapes
            result = util.get_model_dummy_input(
                model, input_shape, input_key, batchsize
            )
            self.assertEqual(result.keys(), {"audio", "video"})
            for i in range(len(input_key)):
                self.assertEqual(
                    result[input_key[i]].size(), tuple([batchsize] + input_shape[i])
                )
            # input_key is string
            input_key = "video"
            input_shape = [4, 16, 223, 223]
            result = util.get_model_dummy_input(
                model, input_shape, input_key, batchsize
            )
            self.assertEqual(result.keys(), {"video"})
            self.assertEqual(result[input_key].size(), tuple([batchsize] + input_shape))
            # input_key is None
            input_key = None
            input_shape = [4, 16, 223, 223]
            result = util.get_model_dummy_input(
                model, input_shape, input_key, batchsize
            )
            self.assertEqual(result.size(), tuple([batchsize] + input_shape))


class TestUpdateStateFunctions(unittest.TestCase):
    def _compare_states(self, state_1, state_2, check_heads=True):
        compare_states(self, state_1, state_2)

    def _compare_model_state(self, state_1, state_2, check_heads=True):
        return compare_model_state(self, state_1, state_2, check_heads=check_heads)

    def test_update_classy_state(self):
        """
        Tests that the update_classy_state successfully updates from a
        checkpoint
        """
        config = get_fast_test_task_config()
        task = build_task(config)
        task_2 = build_task(config)
        task_2.prepare()
        trainer = LocalTrainer(use_gpu=False)
        trainer.train(task)
        update_classy_state(task_2, task.get_classy_state(deep_copy=True))
        self._compare_states(task.get_classy_state(), task_2.get_classy_state())

    def test_update_classy_model(self):
        """
        Tests that the update_classy_model successfully updates from a
        checkpoint
        """
        config = get_fast_test_task_config()
        task = build_task(config)
        use_gpu = torch.cuda.is_available()
        trainer = LocalTrainer(use_gpu=use_gpu)
        trainer.train(task)
        for reset_heads in [False, True]:
            task_2 = build_task(config)
            # prepare task_2 for the right device
            task_2.prepare(use_gpu=use_gpu)
            update_classy_model(
                task_2.model, task.model.get_classy_state(deep_copy=True), reset_heads
            )
            self._compare_model_state(
                task.model.get_classy_state(),
                task_2.model.get_classy_state(),
                check_heads=not reset_heads,
            )
            if reset_heads:
                # the model head states should be different
                with self.assertRaises(Exception):
                    self._compare_model_state(
                        task.model.get_classy_state(),
                        task_2.model.get_classy_state(),
                        check_heads=True,
                    )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

import torch
from classy_vision.losses import ClassyLoss, SumArbitraryLoss, build_loss, register_loss


@register_loss("mock_a")
class MockLoss1(ClassyLoss):
    def forward(self, pred, target):
        return torch.tensor(1.0)

    @classmethod
    def from_config(cls, config):
        return cls()


@register_loss("mock_b")
class MockLoss2(ClassyLoss):
    def forward(self, pred, target):
        return torch.tensor(2.0)

    @classmethod
    def from_config(cls, config):
        return cls()


@register_loss("mock_c")
class MockLoss3(ClassyLoss):
    def forward(self, pred, target):
        return torch.tensor(3.0)

    @classmethod
    def from_config(cls, config):
        return cls()


class TestSumArbitraryLoss(unittest.TestCase):
    def _get_config(self):
        return {
            "name": "sum_arbitrary",
            "weights": [1.0, 1.0, 1.0],
            "losses": [{"name": "mock_a"}, {"name": "mock_b"}, {"name": "mock_c"}],
        }

    def _get_outputs(self):
        return torch.tensor([[2.0, 8.0]])

    def _get_targets(self):
        return torch.tensor([1])

    def test_build_sum_arbitrary(self):
        config = self._get_config()
        crit = build_loss(config)
        self.assertTrue(isinstance(crit, SumArbitraryLoss))
        self.assertAlmostEqual(crit.weights, [1.0, 1.0, 1.0])
        mod_list = [MockLoss1, MockLoss2, MockLoss3]
        for idx, crit_type in enumerate(mod_list):
            self.assertTrue(isinstance(crit.losses[idx], crit_type))

    def test_sum_arbitrary(self):
        config = self._get_config()
        crit = SumArbitraryLoss.from_config(config)
        outputs = self._get_outputs()
        targets = self._get_targets()
        self.assertAlmostEqual(crit(outputs, targets).item(), 1.0 + 2.0 + 3.0)

        # Verify changing losses works
        new_config = copy.deepcopy(config)
        new_config.update(
            {"losses": [{"name": "mock_a"}, {"name": "mock_b"}], "weights": [1.0, 1.0]}
        )
        crit = SumArbitraryLoss.from_config(new_config)
        self.assertAlmostEqual(crit(outputs, targets).item(), 1.0 + 2.0)

        # Verify changing weights works
        new_config = copy.deepcopy(config)
        new_config.update({"weights": [1.0, 2.0, 3.0]})
        crit = SumArbitraryLoss.from_config(new_config)
        self.assertAlmostEqual(
            crit(outputs, targets).item(), 1.0 + 2.0 * 2.0 + 3.0 * 3.0
        )

    def test_deep_copy(self):
        config = self._get_config()
        crit1 = build_loss(config)
        self.assertTrue(isinstance(crit1, SumArbitraryLoss))
        outputs = self._get_outputs()
        targets = self._get_targets()
        crit1(outputs, targets)

        crit2 = copy.deepcopy(crit1)
        self.assertAlmostEqual(
            crit1(outputs, targets).item(), crit2(outputs, targets).item()
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

import torch
from classy_vision.losses import SoftTargetCrossEntropyLoss, build_loss


class TestSoftTargetCrossEntropyLoss(unittest.TestCase):
    def _get_config(self):
        return {
            "name": "soft_target_cross_entropy",
            "ignore_index": -1,
            "reduction": "mean",
        }

    def _get_outputs(self):
        return torch.tensor([[1.0, 7.0, 0.0, 0.0, 2.0]])

    def _get_targets(self):
        return torch.tensor([[1, 0, 0, 0, 1]])

    def _get_loss(self):
        return 5.51097965

    def test_build_soft_target_cross_entropy(self):
        config = self._get_config()
        crit = build_loss(config)
        self.assertTrue(isinstance(crit, SoftTargetCrossEntropyLoss))
        self.assertEqual(crit._ignore_index, -1)
        self.assertEqual(crit._reduction, "mean")

    def test_soft_target_cross_entropy(self):
        config = self._get_config()
        crit = SoftTargetCrossEntropyLoss.from_config(config)
        outputs = self._get_outputs()
        targets = self._get_targets()
        self.assertAlmostEqual(crit(outputs, targets).item(), self._get_loss())

        # Verify ignore index works
        outputs = self._get_outputs()
        targets = torch.tensor([[-1, 0, 0, 0, 1]])
        self.assertAlmostEqual(crit(outputs, targets).item(), 5.01097918)

    def test_unnormalized_soft_target_cross_entropy(self):
        config = {
            "name": "soft_target_cross_entropy",
            "ignore_index": -1,
            "reduction": "mean",
            "normalize_targets": None,
        }
        crit = SoftTargetCrossEntropyLoss.from_config(config)
        outputs = self._get_outputs()
        targets = self._get_targets()
        self.assertAlmostEqual(crit(outputs, targets).item(), 11.0219593)

        # Verify ignore index works
        outputs = self._get_outputs()
        targets = torch.tensor([[-1, 0, 0, 0, 1]])
        self.assertAlmostEqual(crit(outputs, targets).item(), 5.01097965)

    def test_ignore_row(self):
        # If a sample has no valid targets, it should be ignored in the reduction.
        config = self._get_config()
        crit = SoftTargetCrossEntropyLoss.from_config(config)
        outputs = torch.tensor([[1.0, 7.0, 0.0, 0.0, 2.0], [4.0, 2.0, 1.0, 6.0, 0.5]])
        targets = torch.tensor([[1, 0, 0, 0, 1], [-1, -1, -1, -1, -1]])
        self.assertAlmostEqual(crit(outputs, targets).item(), self._get_loss())

    def test_deep_copy(self):
        config = self._get_config()
        crit = build_loss(config)
        self.assertTrue(isinstance(crit, SoftTargetCrossEntropyLoss))
        outputs = self._get_outputs()
        targets = self._get_targets()
        crit(outputs, targets)

        crit2 = copy.deepcopy(crit)
        self.assertAlmostEqual(crit2(outputs, targets).item(), self._get_loss())



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

from classy_vision.optim.param_scheduler import build_param_scheduler
from classy_vision.optim.param_scheduler.composite_scheduler import (
    CompositeParamScheduler,
    UpdateInterval,
)


class TestCompositeScheduler(unittest.TestCase):
    _num_epochs = 10

    def _get_valid_long_config(self):
        return {
            "name": "composite",
            "schedulers": [
                {"name": "constant", "value": 0.1},
                {"name": "constant", "value": 0.2},
                {"name": "constant", "value": 0.3},
                {"name": "constant", "value": 0.4},
            ],
            "lengths": [0.2, 0.4, 0.1, 0.3],
        }

    def _get_lengths_sum_less_one_config(self):
        return {
            "name": "composite",
            "schedulers": [
                {"name": "constant", "value": 0.1},
                {"name": "constant", "value": 0.2},
            ],
            "lengths": [0.7, 0.2999],
        }

    def _get_valid_mixed_config(self):
        return {
            "name": "composite",
            "schedulers": [
                {"name": "step", "values": [0.1, 0.2, 0.3, 0.4, 0.5], "num_epochs": 10},
                {"name": "cosine", "start_lr": 0.42, "end_lr": 0.0001},
            ],
            "lengths": [0.5, 0.5],
        }

    def _get_valid_linear_config(self):
        return {
            "name": "composite",
            "schedulers": [
                {"name": "linear", "start_lr": 0.0, "end_lr": 0.5},
                {"name": "linear", "start_lr": 0.5, "end_lr": 1.0},
            ],
            "lengths": [0.5, 0.5],
            "interval_scaling": ["rescaled", "rescaled"],
        }

    def test_invalid_config(self):
        config = self._get_valid_mixed_config()
        bad_config = copy.deepcopy(config)

        # No schedulers
        bad_config["schedulers"] = []
        bad_config["lengths"] = []
        with self.assertRaises(AssertionError):
            CompositeParamScheduler.from_config(bad_config)

        # Size of schedulers and lengths doesn't match
        bad_config["schedulers"] = copy.deepcopy(config["schedulers"])
        bad_config["lengths"] = copy.deepcopy(config["lengths"])
        bad_config["schedulers"].append(bad_config["schedulers"][-1])
        with self.assertRaises(AssertionError):
            CompositeParamScheduler.from_config(bad_config)

        # Sum of lengths < 1
        bad_config["schedulers"] = copy.deepcopy(config["schedulers"])
        bad_config["lengths"][-1] -= 0.1
        with self.assertRaises(AssertionError):
            CompositeParamScheduler.from_config(bad_config)

        # Sum of lengths > 1
        bad_config["lengths"] = copy.deepcopy(config["lengths"])
        bad_config["lengths"][-1] += 0.1
        with self.assertRaises(AssertionError):
            CompositeParamScheduler.from_config(bad_config)

        # Bad value for update_interval
        bad_config["lengths"] = copy.deepcopy(config["lengths"])
        bad_config["update_interval"] = "epochs"
        with self.assertRaises(AssertionError):
            CompositeParamScheduler.from_config(bad_config)

        # Bad value for composition_mode
        del bad_config["update_interval"]
        bad_config["interval_scaling"] = ["rescaled", "rescaleds"]
        with self.assertRaises(AssertionError):
            CompositeParamScheduler.from_config(bad_config)

        # Wrong number composition modes
        del bad_config["interval_scaling"]
        bad_config["interval_scaling"] = ["rescaled"]
        with self.assertRaises(AssertionError):
            CompositeParamScheduler.from_config(bad_config)

        # Missing required parameters
        del bad_config["interval_scaling"]
        bad_config["lengths"] = config["lengths"]
        del bad_config["lengths"]
        with self.assertRaises(AssertionError):
            CompositeParamScheduler.from_config(bad_config)

        bad_config["lengths"] = config["lengths"]
        del bad_config["schedulers"]
        with self.assertRaises(AssertionError):
            CompositeParamScheduler.from_config(bad_config)

    def test_long_scheduler(self):
        config = self._get_valid_long_config()

        scheduler = CompositeParamScheduler.from_config(config)
        schedule = [
            scheduler(epoch_num / self._num_epochs)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [0.1, 0.1, 0.2, 0.2, 0.2, 0.2, 0.3, 0.4, 0.4, 0.4]

        self.assertEqual(schedule, expected_schedule)

    def test_scheduler_lengths_within_epsilon_of_one(self):
        config = self._get_lengths_sum_less_one_config()
        scheduler = CompositeParamScheduler.from_config(config)
        schedule = [
            scheduler(epoch_num / self._num_epochs)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.2, 0.2, 0.2]
        self.assertEqual(schedule, expected_schedule)

    def test_scheduler_update_interval(self):
        config = self._get_valid_mixed_config()

        # Check default
        scheduler = CompositeParamScheduler.from_config(config)
        self.assertEqual(scheduler.update_interval, UpdateInterval.STEP)

        # Check step
        step_config = copy.deepcopy(config)
        step_config["update_interval"] = "step"
        scheduler = build_param_scheduler(step_config)
        self.assertEqual(scheduler.update_interval, UpdateInterval.STEP)

        # Check epoch
        epoch_config = copy.deepcopy(config)
        epoch_config["update_interval"] = "epoch"
        scheduler = build_param_scheduler(epoch_config)
        self.assertEqual(scheduler.update_interval, UpdateInterval.EPOCH)

    def test_build_composite_scheduler(self):
        config = self._get_valid_mixed_config()
        scheduler = build_param_scheduler(config)
        self.assertTrue(isinstance(scheduler, CompositeParamScheduler))

    def test_scheduler_with_mixed_types(self):
        config = self._get_valid_mixed_config()
        scheduler_0 = build_param_scheduler(config["schedulers"][0])
        scheduler_1 = build_param_scheduler(config["schedulers"][1])

        # Check scaled
        config["interval_scaling"] = ["rescaled", "rescaled"]
        scheduler = CompositeParamScheduler.from_config(config)
        scaled_schedule = [
            round(scheduler(epoch_num / self._num_epochs), 4)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [
            round(scheduler_0(epoch_num / self._num_epochs), 4)
            for epoch_num in range(0, self._num_epochs, 2)
        ] + [
            round(scheduler_1(epoch_num / self._num_epochs), 4)
            for epoch_num in range(0, self._num_epochs, 2)
        ]
        self.assertEqual(scaled_schedule, expected_schedule)

        # Check fixed
        config["interval_scaling"] = ["fixed", "fixed"]
        scheduler = CompositeParamScheduler.from_config(config)
        fixed_schedule = [
            round(scheduler(epoch_num / self._num_epochs), 4)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [
            round(scheduler_0(epoch_num / self._num_epochs), 4)
            for epoch_num in range(0, int(self._num_epochs / 2))
        ] + [
            round(scheduler_1(epoch_num / self._num_epochs), 4)
            for epoch_num in range(int(self._num_epochs / 2), self._num_epochs)
        ]
        self.assertEqual(fixed_schedule, expected_schedule)

        # Check that default is rescaled
        del config["interval_scaling"]
        scheduler = CompositeParamScheduler.from_config(config)
        schedule = [
            round(scheduler(epoch_num / self._num_epochs), 4)
            for epoch_num in range(self._num_epochs)
        ]
        self.assertEqual(scaled_schedule, schedule)
        # Check warmup of rescaled then fixed
        config["interval_scaling"] = ["rescaled", "fixed"]
        scheduler = CompositeParamScheduler.from_config(config)
        fixed_schedule = [
            round(scheduler(epoch_num / self._num_epochs), 4)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [
            round(scheduler_0(epoch_num / self._num_epochs), 4)
            for epoch_num in range(0, int(self._num_epochs), 2)
        ] + [
            round(scheduler_1(epoch_num / self._num_epochs), 4)
            for epoch_num in range(int(self._num_epochs / 2), self._num_epochs)
        ]
        self.assertEqual(fixed_schedule, expected_schedule)

    def test_linear_scheduler_no_gaps(self):
        config = self._get_valid_linear_config()

        # Check rescaled
        scheduler = CompositeParamScheduler.from_config(config)
        schedule = [
            scheduler(epoch_num / self._num_epochs)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
        self.assertEqual(expected_schedule, schedule)

        # Check fixed composition gives same result as only 1 scheduler
        config["schedulers"][1] = config["schedulers"][0]
        config["interval_scaling"] = ["fixed", "fixed"]
        scheduler = CompositeParamScheduler.from_config(config)
        linear_scheduler = build_param_scheduler(config["schedulers"][0])
        schedule = [
            scheduler(epoch_num / self._num_epochs)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [
            linear_scheduler(epoch_num / self._num_epochs)
            for epoch_num in range(self._num_epochs)
        ]
        self.assertEqual(expected_schedule, schedule)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

from classy_vision.generic.registry_utils import import_all_modules

from .classy_task import ClassyTask


FILE_ROOT = Path(__file__).parent


TASK_REGISTRY = {}
TASK_CLASS_NAMES = set()


def build_task(config):
    """Builds a ClassyTask from a config.

    This assumes a 'name' key in the config which is used to determine what
    task class to instantiate. For instance, a config `{"name": "my_task",
    "foo": "bar"}` will find a class that was registered as "my_task"
    (see :func:`register_task`) and call .from_config on it."""

    return TASK_REGISTRY[config["name"]].from_config(config)


def register_task(name):
    """Registers a ClassyTask subclass.

    This decorator allows Classy Vision to instantiate a subclass of ClassyTask
    from a configuration file, even if the class itself is not part of the
    Classy Vision framework. To use it, apply this decorator to a ClassyTask
    subclass, like this:

        @register_task('my_task')
        class MyTask(ClassyTask):
            ...

    To instantiate a task from a configuration file, see :func:`build_task`."""

    def register_task_cls(cls):
        if name in TASK_REGISTRY:
            raise ValueError("Cannot register duplicate task ({})".format(name))
        if not issubclass(cls, ClassyTask):
            raise ValueError(
                "Task ({}: {}) must extend ClassyTask".format(name, cls.__name__)
            )
        if cls.__name__ in TASK_CLASS_NAMES:
            raise ValueError(
                "Cannot register task with duplicate class name ({})".format(
                    cls.__name__
                )
            )
        TASK_REGISTRY[name] = cls
        TASK_CLASS_NAMES.add(cls.__name__)
        return cls

    return register_task_cls


from .classification_task import ClassificationTask  # isort:skip
from .fine_tuning_task import FineTuningTask  # isort:skip

__all__ = [
    "ClassyTask",
    "FineTuningTask",
    "build_task",
    "register_task",
    "ClassificationTask",
]

# automatically import any Python files in the tasks/ directory
import_all_modules(FILE_ROOT, "classy_vision.tasks")



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Callable, Dict, Optional, Union

from classy_vision.dataset import ClassyDataset, register_dataset
from classy_vision.dataset.transforms import ClassyTransform, build_transforms
from torchvision.datasets.cifar import CIFAR10, CIFAR100


class CIFARDataset(ClassyDataset):
    _CIFAR_TYPE = None

    def __init__(
        self,
        split: Optional[str],
        batchsize_per_replica: int,
        shuffle: bool,
        transform: Optional[Union[ClassyTransform, Callable]],
        num_samples: Optional[int],
        root: str,
        download: bool = None,
    ):
        assert self._CIFAR_TYPE in [
            "cifar10",
            "cifar100",
        ], "CIFARDataset must be subclassed and a valid _CIFAR_TYPE provided"
        if self._CIFAR_TYPE == "cifar10":
            dataset = CIFAR10(root=root, train=(split == "train"), download=download)
        if self._CIFAR_TYPE == "cifar100":
            dataset = CIFAR100(root=root, train=(split == "train"), download=download)

        super().__init__(
            dataset, split, batchsize_per_replica, shuffle, transform, num_samples
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "CIFARDataset":
        """Instantiates a CIFARDataset from a configuration.

        Args:
            config: A configuration for a CIFARDataset.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A CIFARDataset instance.
        """
        (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
        ) = cls.parse_config(config)
        split = config.get("split")
        root = config.get("root")
        download = config.get("download")

        transform = build_transforms(transform_config)
        return cls(
            split=split,
            batchsize_per_replica=batchsize_per_replica,
            shuffle=shuffle,
            transform=transform,
            num_samples=num_samples,
            root=root,
            download=download,
        )


@register_dataset("classy_cifar10")
class CIFAR10Dataset(CIFARDataset):
    _CIFAR_TYPE = "cifar10"


@register_dataset("classy_cifar100")
class CIFAR100Dataset(CIFARDataset):
    _CIFAR_TYPE = "cifar100"



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

import torch
from classy_vision.losses import BarronLoss, build_loss


class TestBarronLoss(unittest.TestCase):
    def _get_config(self):
        return {"name": "barron", "size_average": True, "alpha": 1.0, "c": 1.0}

    def _get_outputs(self):
        return torch.tensor([[2.0]])

    def _get_targets(self):
        return torch.tensor([3.0])

    def test_build_barron(self):
        config = self._get_config()
        crit = build_loss(config)
        self.assertTrue(isinstance(crit, BarronLoss))
        self.assertEqual(crit.size_average, config["size_average"])
        self.assertAlmostEqual(crit.alpha, config["alpha"])
        self.assertAlmostEqual(crit.c, config["c"])

    def test_barron(self):
        config = self._get_config()
        crit = BarronLoss.from_config(config)
        outputs = self._get_outputs()
        targets = self._get_targets()
        self.assertAlmostEqual(crit(outputs, targets).item(), 0.41421353816986084)

        # Alpha = 0
        config = self._get_config()
        config["alpha"] = 0.0
        crit = BarronLoss.from_config(config)
        outputs = self._get_outputs()
        targets = self._get_targets()
        self.assertAlmostEqual(crit(outputs, targets).item(), 0.40546512603759766)

        # Alpha = inf
        config = self._get_config()
        config["alpha"] = float("inf")
        crit = BarronLoss.from_config(config)
        outputs = self._get_outputs()
        targets = self._get_targets()
        self.assertAlmostEqual(crit(outputs, targets).item(), 0.39346933364868164)

    def test_deep_copy(self):
        config = self._get_config()
        crit1 = build_loss(config)
        self.assertTrue(isinstance(crit1, BarronLoss))
        outputs = self._get_outputs()
        targets = self._get_targets()
        crit1(outputs, targets)

        crit2 = copy.deepcopy(crit1)
        self.assertAlmostEqual(
            crit1(outputs, targets).item(), crit2(outputs, targets).item()
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .list_dataset import ListDataset
from .random_image_datasets import RandomImageBinaryClassDataset, RandomImageDataset
from .random_video_datasets import RandomVideoDataset


__all__ = [
    "ListDataset",
    "RandomImageBinaryClassDataset",
    "RandomImageDataset",
    "RandomVideoDataset",
]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

from classy_vision.generic.registry_utils import import_all_modules

from .classy_optimizer import ClassyOptimizer


FILE_ROOT = Path(__file__).parent


OPTIMIZER_REGISTRY = {}
OPTIMIZER_CLASS_NAMES = set()


def build_optimizer(config):
    """Builds a ClassyOptimizer from a config.

    This assumes a 'name' key in the config which is used to determine what
    optimizer class to instantiate. For instance, a config `{"name": "my_optimizer",
    "foo": "bar"}` will find a class that was registered as "my_optimizer"
    (see :func:`register_optimizer`) and call .from_config on it."""
    return OPTIMIZER_REGISTRY[config["name"]].from_config(config)


def register_optimizer(name):
    """Registers a ClassyOptimizer subclass.

    This decorator allows Classy Vision to instantiate a subclass of
    ClassyOptimizer from a configuration file, even if the class itself is not
    part of the Classy Vision framework. To use it, apply this decorator to a
    ClassyOptimizer subclass, like this:

        @register_optimizer('my_optimizer')
        class MyOptimizer(ClassyOptimizer):
            ...

    To instantiate an optimizer from a configuration file, see
    :func:`build_optimizer`."""

    def register_optimizer_cls(cls):
        if name in OPTIMIZER_REGISTRY:
            raise ValueError("Cannot register duplicate optimizer ({})".format(name))
        if not issubclass(cls, ClassyOptimizer):
            raise ValueError(
                "Optimizer ({}: {}) must extend ClassyVisionOptimizer".format(
                    name, cls.__name__
                )
            )
        if cls.__name__ in OPTIMIZER_CLASS_NAMES:
            raise ValueError(
                "Cannot register optimizer with duplicate class name({})".format(
                    cls.__name__
                )
            )
        OPTIMIZER_REGISTRY[name] = cls
        OPTIMIZER_CLASS_NAMES.add(cls.__name__)
        return cls

    return register_optimizer_cls


# automatically import any Python files in the optim/ directory
import_all_modules(FILE_ROOT, "classy_vision.optim")

from .rmsprop import RMSProp  # isort:skip
from .sgd import SGD  # isort:skip

__all__ = ["ClassyOptimizer", "RMSProp", "SGD"]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from collections import defaultdict, deque
from time import perf_counter
from typing import List, Mapping, Optional, Tuple

import torch
from torch.cuda import Event as CudaEvent


class PerfTimer:
    """
    Very simple timing wrapper, with context manager wrapping.
    Typical usage:

      with PerfTimer('forward_pass', perf_stats):
          model.forward(data)
      # ...
      with PerfTimer('backward_pass', perf_stats):
          model.backward(loss)
      # ...
      print(perf_stats.report_str())

    Note that timer stats accumulate by name, so you can as if resume them
    by re-using the name.

    You can also use it without context manager, i.e. via start() / stop() directly.

    If supplied PerfStats is constructed with use_cuda_events=True (which is default),
    then Cuda events will be added to correctly track time of async execution
    of Cuda kernels:

      with PerfTimer('foobar', perf_stats):
          some_cpu_work()
          schedule_some_cuda_work()

    In example above, the "Host" column will capture elapsed time from the perspective
    of the Python process, and "CudaEvent" column will capture elapsed time between
    scheduling of Cuda work (within the PerfTimer scope) and completion of this work,
    some of which might happen outside the PerfTimer scope.

    If perf_stats is None, using PerfTimer does nothing.
    """

    def __init__(self, timer_name: str, perf_stats: Optional["PerfStats"]):
        self.skip: bool = False
        if perf_stats is None:
            self.skip = True
            return

        self.name: str = timer_name
        self.elapsed: float = 0.0

        self._last_interval: float = 0.0
        self._perf_stats: PerfStats = perf_stats
        self._is_running: bool = False

        if perf_stats.use_cuda_events():
            self._cuda_event_intervals: List[Tuple[CudaEvent, CudaEvent]] = []

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exception, traceback):
        self.stop()
        if exc_type is None:
            # Only record timer value if with-context finished without error
            self.record()

        return False  # re-raise if there was exception

    def start(self):
        if self.skip or self._is_running:
            return

        self._last_interval = 0.0
        self._is_running = True
        self._start_time: float = perf_counter()
        if self._perf_stats.use_cuda_events():
            self._start_event = torch.cuda.Event(enable_timing=True)
            self._start_event.record()

    def stop(self):
        if self.skip or not self._is_running:
            return

        self._last_interval = perf_counter() - self._start_time
        self.elapsed += self._last_interval

        if self._perf_stats.use_cuda_events():
            # Two cuda events will measure real GPU time within PerfTimer scope:
            end_event = torch.cuda.Event(enable_timing=True)
            end_event.record()
            self._cuda_event_intervals.append((self._start_event, end_event))

        self._is_running = False

    def record(self):
        if self.skip:
            return
        assert not self._is_running
        self._perf_stats.update_with_timer(self)


class PerfMetric:
    """
    Encapsulates numerical tracking of a single metric, with a `.update(value)` API.
    Under-the-hood this can additionally keep track of sums, (exp.) moving averages,
    sum of squares (e.g. for stdev), filtered values, etc.
    """

    # Coefficient for exponential moving average (EMA).
    # Value of 0.1 means last 8 values account for ~50% of weight.
    EMA_FACTOR = 0.1

    def __init__(self):
        self.last_value: Optional[float] = None
        self.smoothed_value: Optional[float] = None

        self.sum_values: float = 0.0
        self.num_updates: int = 0

    def update(self, value: float):
        self.last_value = value
        if self.smoothed_value is None:
            self.smoothed_value = value
        else:
            # TODO (T47970762): correct for initialization bias
            self.smoothed_value = (
                PerfMetric.EMA_FACTOR * value
                + (1.0 - PerfMetric.EMA_FACTOR) * self.smoothed_value
            )

        self.sum_values += value
        self.num_updates += 1

    def get_avg(self):
        if self.num_updates == 0:
            return 0.0
        else:
            return self.sum_values / self.num_updates


class PerfStats:
    """
    Accumulate stats (from timers) over many iterations
    """

    MAX_PENDING_TIMERS = 1000

    def __init__(self, use_cuda_events=True):
        self._host_stats: Mapping[str, PerfMetric] = defaultdict(PerfMetric)
        self._cuda_stats: Mapping[str, PerfMetric] = defaultdict(PerfMetric)

        if use_cuda_events:
            if torch.cuda.is_available():
                self._cuda_pending_timers = deque(maxlen=PerfStats.MAX_PENDING_TIMERS)
            else:
                logging.warning("CUDA unavailable: CUDA events are not logged.")
                self._cuda_pending_timers = None
        else:
            self._cuda_pending_timers = None

    def update_with_timer(self, timer: PerfTimer):
        self._host_stats[timer.name].update(timer._last_interval)

        if self.use_cuda_events():
            if len(self._cuda_pending_timers) >= self._cuda_pending_timers.maxlen:
                logging.error(
                    "Too many pending timers. CudaEvent-based stats will be inaccurate!"
                )
            else:
                self._cuda_pending_timers.append(timer)
            self._process_cuda_events()

    def _process_cuda_events(self):
        """
        Service pending timers. Dequeue timers and aggregate Cuda time intervals,
        until the first "pending" timer (i.e. dependent on a not-yet-ready cuda event).
        """
        while len(self._cuda_pending_timers) > 0:
            timer = self._cuda_pending_timers[0]
            elapsed_cuda = 0.0

            for ev_start, ev_end in timer._cuda_event_intervals:
                if not ev_start.query() or not ev_end.query():
                    # Cuda events associated with this timer aren't ready yet,
                    # stop servicing the queue.
                    return
                # Use seconds (instead of ms) for consistency with "host" timers
                elapsed_cuda += ev_start.elapsed_time(ev_end) / 1000.0

            # All time intervals for this timer are now accounted for.
            # Aggregate stats and pop from pending queue.
            self._cuda_stats[timer.name].update(elapsed_cuda)
            self._cuda_pending_timers.popleft()

    def report_str(self):
        """
        Fancy column-aligned human-readable report.
        If using Cuda events, calling this invokes cuda.synchronize(), which is needed
        to capture pending Cuda work in the report.
        """
        if self.use_cuda_events():
            torch.cuda.synchronize()
            self._process_cuda_events()

        name_width = max(len(k) for k in self._host_stats.keys())

        header = ("{:>" + str(name_width + 4) + "s}  {:>7s}    {:>7s}").format(
            "Timer", "Host", "CudaEvent"
        )
        row_fmt = "{:>" + str(name_width + 4) + "s}: {:>7.2f} ms {:>7.2f} ms"

        rows = []
        rows.append(header)
        for name, metric in self._host_stats.items():
            rows.append(
                row_fmt.format(
                    name,
                    metric.get_avg() * 1000.0,
                    self._cuda_stats[name].get_avg() * 1000.0,
                )
            )
        return "\n".join(rows)

    def use_cuda_events(self):
        return torch.cuda.is_available() and self._cuda_pending_timers is not None

    def __str__(self):
        return str((self._host_stats, self._cuda_stats))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any, Dict

from classy_vision import tasks
from classy_vision.generic.profiler import compute_flops, count_params
from classy_vision.hooks.classy_hook import ClassyHook


class ModelComplexityHook(ClassyHook):
    """
    Logs the number of paramaters and forward pass FLOPs of the model.
    """

    on_rendezvous = ClassyHook._noop
    on_phase_start = ClassyHook._noop
    on_sample = ClassyHook._noop
    on_forward = ClassyHook._noop
    on_loss_and_meter = ClassyHook._noop
    on_backward = ClassyHook._noop
    on_update = ClassyHook._noop
    on_phase_end = ClassyHook._noop
    on_end = ClassyHook._noop

    def on_start(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Measure number of parameters and number of FLOPs."""
        try:
            num_flops = compute_flops(
                task.base_model,
                input_shape=task.base_model.input_shape,
                input_key=task.base_model.input_key
                if hasattr(task.base_model, "input_key")
                else None,
            )
            if num_flops is None:
                logging.info("FLOPs for forward pass: skipped.")
            else:
                logging.info(
                    "FLOPs for forward pass: %d MFLOPs" % (float(num_flops) / 1e6)
                )
        except NotImplementedError:
            logging.warning(
                """Model contains unsupported modules:
            Could not compute FLOPs for model forward pass. Exception:""",
                exc_info=True,
            )
        logging.info(
            "Number of parameters in model: %d" % count_params(task.base_model)
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from test.generic.meter_test_utils import ClassificationMeterTest

import torch
from classy_vision import meters
from classy_vision.meters import AccuracyMeter


class TestAccuracyMeter(ClassificationMeterTest):
    def test_accuracy_meter_registry(self):
        accuracy_meter = meters.build_meter({"name": "accuracy", "topk": [1, 2]})
        self.assertTrue(isinstance(accuracy_meter, AccuracyMeter))

    def test_single_meter_update_and_reset(self):
        """
        This test verifies that the meter works as expected on a single
        update + reset + same single update.
        """
        meter = AccuracyMeter(topk=[1, 2])

        # Batchsize = 3, num classes = 3, score is a value in {1, 2,
        # 3}...3 is the highest score
        model_output = torch.tensor([[3, 2, 1], [3, 1, 2], [1, 3, 2]])

        # Class 0 is the correct class for sample 1, class 2 for sample 2, etc
        target = torch.tensor([0, 1, 2])

        # Only the first sample has top class correct, first and third
        # sample have correct class in top 2
        expected_value = {"top_1": 1 / 3.0, "top_2": 2 / 3.0}

        self.meter_update_and_reset_test(meter, model_output, target, expected_value)

    def test_double_meter_update_and_reset(self):
        meter = AccuracyMeter(topk=[1, 2])

        # Batchsize = 3, num classes = 3, score is a value in {1, 2,
        # 3}...3 is the highest score...two batches in this test
        model_outputs = [
            torch.tensor([[3, 2, 1], [3, 1, 2], [1, 3, 2]]),
            torch.tensor([[3, 2, 1], [1, 3, 2], [1, 3, 2]]),
        ]

        # Class 0 is the correct class for sample 1, class 2 for
        # sample 2, etc, in both batches
        targets = [torch.tensor([0, 1, 2]), torch.tensor([0, 1, 2])]

        # First batch has top-1 accuracy of 1/3.0, top-2 accuracy of 2/3.0
        # Second batch has top-1 accuracy of 2/3.0, top-2 accuracy of 3/3.0
        expected_value = {"top_1": 3 / 6.0, "top_2": 5 / 6.0}

        self.meter_update_and_reset_test(meter, model_outputs, targets, expected_value)

    def test_meter_invalid_model_output(self):
        meter = AccuracyMeter(topk=[1, 2])
        # This model output has 3 dimensions instead of expected 2
        model_output = torch.tensor(
            [[[3, 2, 1], [1, 2, 3]], [[-1, -3, -4], [-10, -90, -100]]]
        )
        target = torch.tensor([0, 1, 2])

        self.meter_invalid_meter_input_test(meter, model_output, target)

    def test_meter_invalid_target(self):
        meter = AccuracyMeter(topk=[1, 2])
        model_output = torch.tensor([[3, 2, 1], [3, 1, 2], [1, 3, 2]])
        # Target has 2 dimensions instead of expected 1
        target = torch.tensor([[0, 1, 2], [0, 1, 2]])

        self.meter_invalid_meter_input_test(meter, model_output, target)

    def test_meter_invalid_topk(self):
        meter = AccuracyMeter(topk=[1, 5])
        model_output = torch.tensor([[3, 2, 1], [3, 1, 2], [1, 3, 2]])
        target = torch.tensor([0, 1, 2])

        self.meter_invalid_meter_input_test(meter, model_output, target)

    def test_meter_get_set_classy_state_test(self):
        # In this test we update meter0 with model_output0 & target0
        # and we update meter1 with model_output1 & target1 then
        # transfer the state from meter1 to meter0 and validate they
        # give same expected value.
        # Expected value is the expected value of meter1
        meters = [AccuracyMeter(topk=[1, 2]), AccuracyMeter(topk=[1, 2])]

        # Batchsize = 3, num classes = 3, score is a value in {1, 2,
        # 3}...3 is the highest score
        model_outputs = [
            torch.tensor([[1, 2, 3], [1, 2, 3], [2, 3, 1]]),
            torch.tensor([[3, 2, 1], [3, 1, 2], [1, 3, 2]]),
        ]

        # Class 0 is the correct class for sample 1, class 2 for sample 2, etc
        targets = [torch.tensor([0, 1, 2]), torch.tensor([0, 1, 2])]

        # Value for second update
        expected_value = {"top_1": 1 / 3.0, "top_2": 2 / 3.0}

        self.meter_get_set_classy_state_test(
            meters, model_outputs, targets, expected_value
        )

    def test_meter_distributed(self):
        # Meter0 will execute on one process, Meter1 on the other
        meters = [AccuracyMeter(topk=[1, 2]), AccuracyMeter(topk=[1, 2])]

        # Batchsize = 3, num classes = 3, score is a value in {1, 2,
        # 3}...3 is the highest score
        model_outputs = [
            torch.tensor([[3, 2, 1], [3, 1, 2], [1, 3, 2]]),  # Meter 0
            torch.tensor([[3, 2, 1], [1, 3, 2], [1, 3, 2]]),  # Meter 1
            torch.tensor([[3, 2, 1], [3, 1, 2], [1, 3, 2]]),  # Meter 0
            torch.tensor([[3, 2, 1], [1, 3, 2], [1, 3, 2]]),  # Meter 1
        ]

        # Class 0 is the correct class for sample 1, class 2 for sample 2, etc
        targets = [
            torch.tensor([0, 1, 2]),  # Meter 0
            torch.tensor([0, 1, 2]),  # Meter 1
            torch.tensor([0, 1, 2]),  # Meter 0
            torch.tensor([0, 1, 2]),  # Meter 1
        ]

        # In first two updates there are 3 correct top-2, 5 correct in top 2
        # The same occurs in the second two updates and is added to first
        expected_values = [
            {"top_1": 3 / 6.0, "top_2": 5 / 6.0},  # After one update to each meter
            {"top_1": 6 / 12.0, "top_2": 10 / 12.0},  # After two updates to each meter
        ]

        self.meter_distributed_test(meters, model_outputs, targets, expected_values)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# TODO: Some batch-normalization operations are superfluous and can be removed.

# dependencies:
import math
from typing import Any, Dict

import torch
import torch.nn as nn
from classy_vision.generic.util import is_pos_int

from . import register_model
from .classy_model import ClassyModel


# global setting for in-place ReLU:
INPLACE = True


class _DenseLayer(nn.Sequential):
    """
        Single layer of a DenseNet.
    """

    def __init__(self, in_planes, growth_rate=32, expansion=4):

        # assertions:
        assert is_pos_int(in_planes)
        assert is_pos_int(growth_rate)
        assert is_pos_int(expansion)

        # add all layers to layer
        super(_DenseLayer, self).__init__()
        intermediate = expansion * growth_rate
        self.add_module("norm-1", nn.BatchNorm2d(in_planes))
        self.add_module("relu-1", nn.ReLU(inplace=INPLACE))
        self.add_module(
            "conv-1",
            nn.Conv2d(in_planes, intermediate, kernel_size=1, stride=1, bias=False),
        )
        self.add_module("norm-2", nn.BatchNorm2d(intermediate))
        self.add_module("relu-2", nn.ReLU(inplace=INPLACE))
        self.add_module(
            "conv-2",
            nn.Conv2d(
                intermediate,
                growth_rate,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

    def forward(self, x):
        new_features = super(_DenseLayer, self).forward(x)
        return torch.cat([x, new_features], 1)


class _DenseBlock(nn.Sequential):
    """
        Block of densely connected layers at same resolution.
    """

    def __init__(self, num_layers, in_planes, growth_rate=32, expansion=4):

        # assertions:
        assert is_pos_int(in_planes)
        assert is_pos_int(growth_rate)
        assert is_pos_int(expansion)

        # create block of dense layers at same resolution:
        super(_DenseBlock, self).__init__()
        for idx in range(num_layers):
            layer = _DenseLayer(
                in_planes + idx * growth_rate,
                growth_rate=growth_rate,
                expansion=expansion,
            )
            self.add_module("denselayer-%d" % (idx + 1), layer)


class _Transition(nn.Sequential):
    """
        Transition layer to reduce spatial resolution.
    """

    def __init__(self, in_planes, out_planes, reduction=2):

        # assertions:
        assert is_pos_int(in_planes)
        assert is_pos_int(out_planes)
        assert is_pos_int(reduction)

        # create layers for pooling:
        super(_Transition, self).__init__()
        self.add_module("pool-norm", nn.BatchNorm2d(in_planes))
        self.add_module("pool-relu", nn.ReLU(inplace=INPLACE))
        self.add_module(
            "pool-conv",
            nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=1, bias=False),
        )
        self.add_module(
            "pool-pool", nn.AvgPool2d(kernel_size=reduction, stride=reduction)
        )


@register_model("densenet")
class DenseNet(ClassyModel):
    def __init__(
        self,
        num_blocks,
        num_classes,
        init_planes,
        growth_rate,
        expansion,
        small_input,
        final_bn_relu,
    ):
        """
            Implementation of a standard densely connected network (DenseNet).

            Set `small_input` to `True` for 32x32 sized image inputs.

            Set `final_bn_relu` to `False` to exclude the final batchnorm and ReLU
            layers. These settings are useful when
            training Siamese networks.
        """
        super().__init__()

        # assertions:
        assert type(num_blocks) == list
        assert all(is_pos_int(b) for b in num_blocks)
        assert num_classes is None or is_pos_int(num_classes)
        assert is_pos_int(init_planes)
        assert is_pos_int(growth_rate)
        assert is_pos_int(expansion)
        assert type(small_input) == bool

        # initial convolutional block:
        self._num_classes = num_classes
        self.num_blocks = num_blocks
        self.small_input = small_input
        if self.small_input:
            self.initial_block = nn.Sequential(
                nn.Conv2d(
                    3, init_planes, kernel_size=3, stride=1, padding=1, bias=False
                ),
                nn.BatchNorm2d(init_planes),
                nn.ReLU(inplace=INPLACE),
            )
        else:
            self.initial_block = nn.Sequential(
                nn.Conv2d(
                    3, init_planes, kernel_size=7, stride=2, padding=3, bias=False
                ),
                nn.BatchNorm2d(init_planes),
                nn.ReLU(inplace=INPLACE),
                nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
            )
        # loop over spatial resolutions:
        num_planes = init_planes
        self.features = nn.Sequential()
        for idx, num_layers in enumerate(num_blocks):

            # add dense block:
            block = _DenseBlock(
                num_layers, num_planes, growth_rate=growth_rate, expansion=expansion
            )
            self.features.add_module("denseblock-%d" % (idx + 1), block)
            num_planes = num_planes + num_layers * growth_rate

            # add transition layer:
            if idx != len(num_blocks) - 1:
                trans = _Transition(num_planes, num_planes // 2)
                self.features.add_module("transition-%d" % (idx + 1), trans)
                num_planes = num_planes // 2

        # final batch normalization:
        if final_bn_relu:
            self.features.add_module("norm-final", nn.BatchNorm2d(num_planes))
            self.features.add_module("relu-final", nn.ReLU(inplace=INPLACE))

        # final classifier:
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = None if num_classes is None else nn.Linear(num_planes, num_classes)
        self.num_planes = num_planes

        # initialize weights of convolutional and batchnorm layers:
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                m.bias.data.zero_()

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "DenseNet":
        """Instantiates a DenseNet from a configuration.

        Args:
            config: A configuration for a DenseNet.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A DenseNet instance.
        """
        assert "num_blocks" in config
        config = {
            "num_blocks": config["num_blocks"],
            "num_classes": config.get("num_classes"),
            "init_planes": config.get("init_planes", 64),
            "growth_rate": config.get("growth_rate", 32),
            "expansion": config.get("expansion", 4),
            "small_input": config.get("small_input", False),
            "final_bn_relu": config.get("final_bn_relu", True),
        }
        return cls(**config)

    # forward pass in DenseNet:
    def forward(self, x):

        # initial convolutional block:
        out = self.initial_block(x)

        # evaluate all dense blocks:
        out = self.features(out)

        # perform average pooling:
        out = self.avgpool(out)

        # final classifier:
        out = out.view(out.size(0), -1)
        if self.fc is not None:
            out = self.fc(out)
        return out

    def get_optimizer_params(self):
        # use weight decay on BatchNorm for DenseNets
        return super().get_optimizer_params(bn_weight_decay=True)

    @property
    def input_shape(self):
        if self.small_input:
            return (3, 32, 32)
        else:
            return (3, 224, 224)

    @property
    def output_shape(self):
        return (1, self._num_classes)

    @property
    def model_depth(self):
        return sum(self.num_blocks)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch

from . import ClassyLoss, build_loss, register_loss


@register_loss("multi_output_sum_loss")
class MultiOutputSumLoss(ClassyLoss):
    """
    Applies the provided loss to the list of outputs (or single output) and sums
    up the losses.
    """

    def __init__(self, loss) -> None:
        super().__init__()

        self._loss = loss

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "MultiOutputSumLoss":
        """Instantiates a MultiOutputSumLoss from a configuration.

        Args:
            config: A configuration for a MultiOutpuSumLoss.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A MultiOutputSumLoss instance.
        """
        assert (
            type(config["loss"]) == dict
        ), "loss must be a dict containing a configuration for a registered loss"
        return cls(loss=build_loss(config["loss"]))

    def forward(self, output, target):
        if torch.is_tensor(output):
            output = [output]

        loss = 0
        for pred in output:
            loss += self._loss(pred, target)

        return loss



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

from classy_vision.optim.param_scheduler import build_param_scheduler
from classy_vision.optim.param_scheduler.linear_scheduler import LinearParamScheduler


class TestLienarScheduler(unittest.TestCase):
    _num_epochs = 10

    def _get_valid_intermediate(self):
        return [0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09]

    def _get_valid_config(self):
        return {"name": "linear", "start_lr": 0.0, "end_lr": 0.1}

    def test_invalid_config(self):
        config = self._get_valid_config()

        bad_config = copy.deepcopy(config)
        # No start lr
        del bad_config["start_lr"]
        with self.assertRaises(AssertionError):
            LinearParamScheduler.from_config(bad_config)

        # No end lr
        bad_config["start_lr"] = config["start_lr"]
        del bad_config["end_lr"]
        with self.assertRaises(AssertionError):
            LinearParamScheduler.from_config(bad_config)

    def test_scheduler(self):
        config = self._get_valid_config()

        # Check as warmup
        scheduler = LinearParamScheduler.from_config(config)
        schedule = [
            round(scheduler(epoch_num / self._num_epochs), 4)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [config["start_lr"]] + self._get_valid_intermediate()
        self.assertEqual(schedule, expected_schedule)

        # Check as decay
        tmp = config["start_lr"]
        config["start_lr"] = config["end_lr"]
        config["end_lr"] = tmp
        scheduler = LinearParamScheduler.from_config(config)
        schedule = [
            round(scheduler(epoch_num / self._num_epochs), 4)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [config["start_lr"]] + list(
            reversed(self._get_valid_intermediate())
        )
        self.assertEqual(schedule, expected_schedule)

    def test_build_linear_scheduler(self):
        config = self._get_valid_config()
        scheduler = build_param_scheduler(config)
        self.assertTrue(isinstance(scheduler, LinearParamScheduler))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Implementation of ResNet (https://arxiv.org/pdf/1512.03385.pdf) as a special
case of ResNeXt (https://arxiv.org/pdf/1611.05431.pdf)
"""

from . import register_model
from .resnext import ResNeXt


# global setting for in-place ReLU:
INPLACE = True


@register_model("resnet")
class ResNet(ResNeXt):
    """
        ResNet is a special case of ResNeXt.
    """

    def __init__(self, **kwargs):
        assert (
            kwargs["base_width_and_cardinality"] is None
        ), "base_width_and_cardinality should be None for ResNet"
        super().__init__(**kwargs)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Callable, Dict, Iterator, List, Optional, Union

import torch
import torch.nn as nn
from classy_vision.dataset import ClassyDataset
from classy_vision.dataset.image_path_dataset import ImagePathDataset
from classy_vision.dataset.transforms.util import build_field_transform_default_imagenet
from classy_vision.models import ClassyModel, ClassyModelWrapper
from classy_vision.tasks import ClassyTask


class ClassyHubInterface:
    """PyTorch Hub interface for classy vision tasks and models.

    The task is optional, but a model is guaranteed to be present.  Do
    not use the constructor directly, instead Use from_task() or
    from_model() to instantiate the class.

    See the examples folder for an example of how to use this class

    Attributes:
        task: If present, task that can be used to train the torchhub model
        model: torchub model

    """

    def __init__(
        self, task: Optional[ClassyTask] = None, model: Optional[ClassyModel] = None
    ) -> None:
        """Constructor for ClassyHubInterface.

        Only one of task or model can be specified at construction
        time. If task is specified then task.model is used to populate
        the model attribute.

        Do not use the constructor directly, instead use from_task()
        or from_model() to instantiate the class.

        Args:
            task: task that can be used to train torchhub model,
                task.model is used to populate the model attribute
            model: torchhub model
        """
        self.task = task
        if task is None:
            assert model is not None, "Need to specify a model if task is None"
            self.model = model
        else:
            assert model is None, "Cannot pass a model if task is not None"
            self.model = task.model

    @classmethod
    def from_task(cls, task: ClassyTask) -> "ClassyHubInterface":
        """Instantiates the ClassyHubInterface from a task.

        This function returns a hub interface based on a ClassyTask.

        Args:
            task: ClassyTask that contains hub model

        """
        return cls(task=task)

    @classmethod
    def from_model(cls, model: Union[nn.Module, ClassyModel]) -> "ClassyHubInterface":
        """Instantiates the ClassyHubInterface from a model.

        This function returns a hub interface based on a ClassyModel

        Args:
            model: torchhub model

        """
        if not isinstance(model, ClassyModel):
            model = ClassyModelWrapper(model)
        return cls(model=model)

    def create_image_dataset(
        self,
        image_paths: Union[List[str], str],
        targets: Optional[List[Any]] = None,
        batchsize_per_replica: int = 32,
        shuffle: bool = True,
        transform: Optional[Callable] = None,
        num_samples: Optional[int] = None,
        phase_type: str = "train",
    ) -> ClassyDataset:
        """Create a ClassyDataset which reads images from image_paths.

        Args:
            image_paths: Can be
                - A single directory location, in which case the data is expected to be
                    arranged in a format similar to
                    `:class:torchvision.datasets.ImageFolder`. The targets will
                    be inferred from the directory structure.
                - A list of paths, in which case the list will contain the paths
                    to all the images. In this situation, the targets can be specified
                    by using the targets argument.
            targets: A list containing the target classes for each image
            batchsize_per_replica: Minibatch size per replica (i.e. samples per GPU)
            shuffle: If true, data is shuffled between epochs
            transform: Transform to apply to sample. If left as None, the dataset's
                phase_type is used to determine the transform to apply. The transform
                for the phase_type is searched for in self.task, falling back to
                imagenet transformations if it is not found there.
            num_samples: If specified, limits the number of samples returned by
                the dataset
            phase_type: String specifying the phase_type, e.g. "train" or "test"
        """
        if transform is None:
            if self.task is not None and phase_type in self.task.datasets:
                # use the transform from the dataset for the phase_type
                dataset = self.task.datasets[phase_type]
                transform = dataset.transform
                assert transform is not None, "Cannot infer transform from the task"
            else:
                transform = build_field_transform_default_imagenet(
                    config=None, split=phase_type
                )
        return ImagePathDataset(
            batchsize_per_replica,
            shuffle,
            transform,
            num_samples,
            image_paths,
            targets=targets,
            split=phase_type,
        )

    @staticmethod
    def get_data_iterator(dataset: ClassyDataset) -> Iterator[Any]:
        """Returns an iterator that can be used to retrieve training / testing samples.

        Args:
            dataset: Dataset to iterate over
        """
        return iter(dataset.iterator())

    def train(self) -> None:
        """Sets the model to train mode and enables torch gradient calculation

        """
        torch.autograd.set_grad_enabled(True)
        self.model.train()

    def eval(self) -> None:
        """Sets the model to eval mode and disables torch gradient calculation

        """
        torch.autograd.set_grad_enabled(False)
        self.model.eval()

    def predict(self, sample):
        """Returns the model's prediction for a sample.

        Args:
            sample: Must contain "input" key, model calculates prediction over input.
        """
        output = self.model(sample["input"])
        # squeeze the output in case the batch size is 1
        return output.squeeze()

    def extract_features(self, sample):
        """Calculates feature embeddings of sample.

        Args:
            sample: Must contain "input" key, model calculates prediction over input.
        """
        output = self.model.extract_features(sample["input"])
        # squeeze the output in case the batch size is 1
        return output.squeeze()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torchvision.transforms as transforms

from . import ClassyTransform, build_transforms, register_transform


class ImagenetConstants:
    """Constant variables related to the image classification.

    MEAN: often used to be subtracted from image RGB value. Computed on ImageNet.
    STD: often used to divide the image RGB value after mean centering. Computed
        on ImageNet.
    CROP_SIZE: the size of image cropping which is often the input to deep network.
    RESIZE: the size of rescaled image.

    """

    MEAN = [0.485, 0.456, 0.406]
    STD = [0.229, 0.224, 0.225]
    CROP_SIZE = 224
    RESIZE = 256


@register_transform("apply_transform_to_key")
class ApplyTransformToKey:
    """Serializable class that applies a transform to a key specified field in samples.
    """

    def __init__(self, transform: Callable, key: Union[int, str] = "input") -> None:
        """The constructor method of ApplyTransformToKey class.

        Args:
            transform: a callable function that takes sample data of type dict as input
            key: the key in sample whose corresponding value will undergo
                the transform

        """
        self.key: Union[int, str] = key
        self.transform: Callable = transform

    @classmethod
    def from_config(cls, config: Dict[str, Any]):
        transform = build_transforms(config["transforms"])

        return cls(transform=transform, key=config["key"])

    def __call__(
        self, sample: Union[Tuple[Any], Dict[str, Any]]
    ) -> Union[Tuple[Any], Dict[str, Any]]:
        """Updates sample by applying a transform to the value at the specified key.

        Args:
            sample: input sample which will be transformed

        """
        if sample is None:
            return sample

        # Asserts + deal with tuple immutability
        convert_to_tuple = False
        if isinstance(sample, dict):
            assert (
                self.key in sample
            ), "This transform only supports dicts with key '{}'".format(self.key)
        elif isinstance(sample, (tuple, list)):
            assert self.key < len(
                sample
            ), "This transform only supports tuples / lists with key less "
            "than {length}, key provided {key}".format(length=len(sample), key=self.key)
            # Convert to list for transformation
            if isinstance(sample, tuple):
                convert_to_tuple = True
            sample = list(sample)

        sample[self.key] = self.transform(sample[self.key])
        if convert_to_tuple:
            sample = tuple(sample)

        return sample


@register_transform("imagenet_augment")
class ImagenetAugmentTransform(ClassyTransform):
    """The default image transform with data augmentation.

    It is often useful for training models on Imagenet. It sequentially resizes
    the image into a random scale, takes a random spatial cropping, randomly flips
    the image horizontally, transforms PIL image data into a torch.Tensor and
    normalizes the pixel values by mean subtraction and standard deviation division.
    """

    def __init__(
        self,
        crop_size: int = ImagenetConstants.CROP_SIZE,
        mean: List[float] = ImagenetConstants.MEAN,
        std: List[float] = ImagenetConstants.STD,
    ):
        """The constructor method of ImagenetAugmentTransform class.

        Args:
            crop_size: expected output size of random cropping
            mean: a 3-tuple denoting the pixel RGB mean
            std: a 3-tuple denoting the pixel RGB standard deviation

        """
        self.transform = transforms.Compose(
            [
                transforms.RandomResizedCrop(crop_size),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize(mean=mean, std=std),
            ]
        )

    def __call__(self, img):
        """Callable function which applies the tranform to the input image.

        Args:
            image: input image that will undergo the transform

        """
        return self.transform(img)


@register_transform("imagenet_no_augment")
class ImagenetNoAugmentTransform(ClassyTransform):
    """The default image transform without data augmentation.

    It is often useful for testing models on Imagenet. It sequentially resizes
    the image, takes a central  cropping, transforms PIL image data into a
    torch.Tensor and normalizes the pixel values by mean subtraction and standard
    deviation division.

    """

    def __init__(
        self,
        resize: int = ImagenetConstants.RESIZE,
        crop_size: int = ImagenetConstants.CROP_SIZE,
        mean: List[float] = ImagenetConstants.MEAN,
        std: List[float] = ImagenetConstants.STD,
    ):
        """The constructor method of ImagenetNoAugmentTransform class.

        Args:
            resize: expected image size after resizing
            crop_size: expected size of central cropping
            mean: a 3-tuple denoting the pixel RGB mean
            std: a 3-tuple denoting the pixel RGB standard deviation

        """
        self.transform = transforms.Compose(
            [
                transforms.Resize(resize),
                transforms.CenterCrop(crop_size),
                transforms.ToTensor(),
                transforms.Normalize(mean=mean, std=std),
            ]
        )

    def __call__(self, img):
        """Callable function which applies the tranform to the input image.

        Args:
            image: input image that will undergo the transform

        """
        return self.transform(img)


@register_transform("generic_image_transform")
class GenericImageTransform(ClassyTransform):
    """Default transform for images used in the classification task

    This transform does several things. First, it expects a tuple or
    list input (torchvision datasets supply tuples / lists). Second,
    it applies a user-provided image transforms to the first entry in
    the tuple (again, matching the torchvision tuple format). Third,
    it transforms the tuple to a dict sample with entries "input" and
    "target".

    The defaults are for the standard imagenet augmentations

    This is just a convenience wrapper to cover the common
    use-case. You can get the same behavior by composing torchvision
    transforms + ApplyTransformToKey + TupleToMap.

    """

    def __init__(
        self, transform: Optional[Callable] = None, split: Optional[str] = None
    ):
        """Constructor for GenericImageTransfrom

        Args:
            transform: A callable or ClassyTransform to be applied to the image only
            split: 'train' or 'test'. Only one of the two arguments
                should be specified
        """
        assert (
            transform is not None or split is not None
        ), "One of transform / split must be specified"
        assert (
            transform is None or split is None
        ), "Only one of transform / split should be specified"
        assert split in [None, "train", "test"], (
            "If specified, split should be either 'train' or 'test', "
            "instead got {}".format(split)
        )

        if transform is not None:
            self._transform = transform

        if split is not None:
            self._transform = (
                ImagenetAugmentTransform()
                if split == "train"
                else ImagenetNoAugmentTransform()
            )

    @classmethod
    def from_config(cls, config: Dict[str, Any]):
        transform = None
        if "transforms" in config:
            transform = build_transforms(config["transforms"])
        split = config.get("split")
        return cls(transform, split)

    def __call__(self, sample: Tuple[Any]):
        """Applied transform to sample

        Args:
            sample: A tuple with length >= 2. The first entry should
                be the image data, the second entry should be the
                target data.
        """
        image = sample[0]
        transformed_image = self._transform(image)
        new_sample = {"input": transformed_image, "target": sample[1]}
        # Any additional metadata is just appended under index of tuple
        if len(sample) > 2:
            for i in range(2, len(sample)):
                new_sample[str(i)] = sample[i]

        return new_sample


@register_transform("tuple_to_map")
class TupleToMapTransform(ClassyTransform):
    """A transform which maps image data from tuple to dict.

    This transform takes a sample of the form (data1, data2, ...) and
    returns a sample of the form {key1: data1, key2: data2, ...}

    It is useful for mapping output from datasets like the PyTorch
    ImageFolder dataset (tuple) to dict with named data fields.

    If sample is already a dict with the required keys, pass sample through.

    """

    def __init__(self, list_of_map_keys: List[str]):
        """The constructor method of TupleToMapTransform class.

        Args:
            list_of_map_keys: a list of dict keys that will be mapped to item
                in the input sample of data type list

        """
        self._map_keys = list_of_map_keys

    def __call__(self, sample):
        """Transform sample from type tuple to type dict.

        Args:
            sample: input sample which will be transformed

        """
        # If already a dict/map with appropriate keys, exit early
        if isinstance(sample, dict):
            for key in self._map_keys:
                assert (
                    key in sample
                ), "Sample {sample} must be a tuple or a dict with keys {keys}".format(
                    sample=str(sample), keys=str(self._map_keys)
                )
            return sample

        assert len(sample) == len(self._map_keys), (
            "Provided sample tuple must have same number of keys "
            "as provided to transform"
        )
        output_sample = {}
        for idx, s in enumerate(sample):
            output_sample[self._map_keys[idx]] = s

        return output_sample


DEFAULT_KEY_MAP = TupleToMapTransform(["input", "target"])


def build_field_transform_default_imagenet(
    config: Optional[List[Dict[str, Any]]],
    default_transform: Optional[Callable] = None,
    split: Optional[bool] = None,
    key: Union[int, str] = "input",
    key_map_transform: Optional[Callable] = DEFAULT_KEY_MAP,
) -> Callable:
    """Returns a ApplyTransformToKey which applies a transform on the specified key.

    The transform is built from the config, if it is not None.

    Otherwise, uses one of the two mutually exclusive args: If
    default_transform is not None, it is used.  If split is not None,
    imagenet transforms are used, using augmentation for "train", no
    augmentation otherwise.

    This function also provides an additional
    function for mapping from tuples (or other keys) to a desired set
    of keys

    Args:
        config: field transform config
        default_transform: used if config is None
        split: split for dataset, e.g. "train" or "test"
        key: Key to apply transform to
        key_map_transform: Used to produce desired map / keys
            (e.g. for torchvision datasets, default samples is a
            tuple so this argument can be used to map
            (input, target) -> {"input": input, "target": target})

    """
    assert (
        default_transform is None or split is None
    ), "Can only specify one of default_transform and split"
    if config is None:
        if default_transform is not None:
            transform = default_transform
        elif split is not None:
            transform = (
                ImagenetAugmentTransform()
                if split == "train"
                else ImagenetNoAugmentTransform()
            )
        else:
            raise ValueError("No transform config provided with no defaults")
    else:
        transform = build_transforms(config)

    transform = ApplyTransformToKey(transform, key=key)
    if key_map_transform is None:
        return transform

    return transforms.Compose([key_map_transform, transform])


def default_unnormalize(img):
    """Default unnormalization transform which undo the "transforms.Normalize".

        Specially, it cancels out mean subtraction and standard deviation division.

    Args:
        img (torch.Tensor): image data to which the transform will be applied

    """
    # TODO T39752655: Allow this to be configurable
    img = img.clone()
    for channel, std, mean in zip(img, ImagenetConstants.STD, ImagenetConstants.MEAN):
        channel.mul_(std).add_(mean)
    return img
